import asyncio
import enum
from datetime import datetime
from multiprocessing import Queue
from unittest.mock import ANY
from uuid import uuid4

import numpy as np
import pytest
from numpy.typing import DTypeLike
from p4p import Value
from p4p.client.asyncio import Context
from p4p.client.thread import Context as ThreadContext
from p4p.nt import NTTable

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller import Controller
from fastcs.datatypes import Bool, Enum, Float, Int, String, Table, Waveform
from fastcs.launch import FastCS
from fastcs.transport.epics.options import EpicsIOCOptions
from fastcs.transport.epics.pva.transport import EpicsPVATransport
from fastcs.wrappers import command


@pytest.mark.asyncio
async def test_ioc(p4p_subprocess: tuple[str, Queue]):
    pv_prefix, _ = p4p_subprocess
    ctxt = Context("pva")

    _parent_pvi = await ctxt.get(f"{pv_prefix}:PVI")
    assert isinstance(_parent_pvi, Value)
    parent_pvi = _parent_pvi.todict()
    assert all(f in parent_pvi for f in ("alarm", "display", "timeStamp", "value"))
    assert parent_pvi["display"] == {"description": "some controller"}
    assert parent_pvi["value"] == {
        "a": {"rw": f"{pv_prefix}:A"},
        "b": {"w": f"{pv_prefix}:B"},
        "child": {
            "d": {
                "v1": f"{pv_prefix}:Child1:PVI",
                "v2": f"{pv_prefix}:Child2:PVI",
            }
        },
        "table": {
            "rw": f"{pv_prefix}:Table",
        },
    }

    child_pvi_pv = parent_pvi["value"]["child"]["d"]["v1"]
    _child_pvi = await ctxt.get(child_pvi_pv)
    assert isinstance(_child_pvi, Value)
    child_pvi = _child_pvi.todict()
    assert all(f in child_pvi for f in ("alarm", "display", "timeStamp", "value"))
    assert child_pvi["display"] == {"description": "some sub controller"}
    assert child_pvi["value"] == {
        "c": {"w": f"{pv_prefix}:Child1:C"},
        "d": {"x": f"{pv_prefix}:Child1:D"},
        "e": {"r": f"{pv_prefix}:Child1:E"},
        "f": {"rw": f"{pv_prefix}:Child1:F"},
        "g": {"rw": f"{pv_prefix}:Child1:G"},
        "h": {"rw": f"{pv_prefix}:Child1:H"},
        "i": {"x": f"{pv_prefix}:Child1:I"},
        "j": {"r": f"{pv_prefix}:Child1:J"},
    }


@pytest.mark.asyncio
async def test_scan_method(p4p_subprocess: tuple[str, Queue]):
    pv_prefix, _ = p4p_subprocess
    ctxt = Context("pva")
    e_values = asyncio.Queue()

    # While the scan method will update every 0.1 seconds, it takes around that
    # time for the p4p transport to update, broadcast, get.
    latency = 1e8

    e_monitor = ctxt.monitor(f"{pv_prefix}:Child1:E", e_values.put)
    try:
        # Throw away the value on the ioc setup so we can compare timestamps
        _ = await e_values.get()

        raw_value = (await e_values.get()).raw
        value = raw_value["value"]
        assert isinstance(value, bool)
        nanoseconds = raw_value["timeStamp"]["nanoseconds"]

        new_raw_value = (await e_values.get()).raw
        assert new_raw_value["value"] is not value
        assert new_raw_value["timeStamp"]["nanoseconds"] == pytest.approx(
            nanoseconds + 1e8, abs=latency
        )
        value = new_raw_value["value"]
        assert isinstance(value, bool)
        nanoseconds = new_raw_value["timeStamp"]["nanoseconds"]

        new_raw_value = (await e_values.get()).raw
        assert new_raw_value["value"] is not value
        assert new_raw_value["timeStamp"]["nanoseconds"] == pytest.approx(
            nanoseconds + 1e8, abs=latency
        )

    finally:
        e_monitor.close()


@pytest.mark.asyncio
async def test_command_method(p4p_subprocess: tuple[str, Queue]):
    pv_prefix, _ = p4p_subprocess
    d_values = asyncio.Queue()
    i_values = asyncio.Queue()
    j_values = asyncio.Queue()
    ctxt = Context("pva")

    d_monitor = ctxt.monitor(f"{pv_prefix}:Child1:D", d_values.put)
    i_monitor = ctxt.monitor(f"{pv_prefix}:Child1:I", i_values.put)
    j_monitor = ctxt.monitor(f"{pv_prefix}:Child1:J", j_values.put)

    try:
        j_initial_value = await j_values.get()
        assert (await d_values.get()).raw.value is False
        await ctxt.put(f"{pv_prefix}:Child1:D", True)
        assert (await d_values.get()).raw.value is True
        # D process hangs for 0.1s, so we wait slightly longer
        await asyncio.sleep(0.2)
        # Value returns to False, signifying completed process
        assert (await d_values.get()).raw.value is False
        # D process increments J by 1
        assert (await j_values.get()).raw.value == j_initial_value + 1

        # First run fails
        before_command_value = (await i_values.get()).raw
        assert before_command_value["value"] is False
        assert before_command_value["alarm"]["severity"] == 0
        assert before_command_value["alarm"]["message"] == ""
        await ctxt.put(f"{pv_prefix}:Child1:I", True)
        assert (await i_values.get()).raw.value is True
        await asyncio.sleep(0.2)

        after_command_value = (await i_values.get()).raw
        assert after_command_value["value"] is False
        assert after_command_value["alarm"]["severity"] == 2
        assert (
            after_command_value["alarm"]["message"] == "I: FAILED WITH THIS WEIRD ERROR"
        )
        # Failed I process does not increment J
        assert j_values.empty()

        # Second run succeeds
        await ctxt.put(f"{pv_prefix}:Child1:I", True)
        assert (await i_values.get()).raw.value is True
        await asyncio.sleep(0.2)
        after_command_value = (await i_values.get()).raw
        # Successful I process increments J by 1
        assert (await j_values.get()).raw.value == j_initial_value + 2

        # On the second run the command succeeded so we left the error state
        assert after_command_value["value"] is False
        assert after_command_value["alarm"]["severity"] == 0
        assert after_command_value["alarm"]["message"] == ""

    finally:
        d_monitor.close()
        i_monitor.close()
        j_monitor.close()


@pytest.mark.asyncio
async def test_numerical_alarms(p4p_subprocess: tuple[str, Queue]):
    pv_prefix, _ = p4p_subprocess
    a_values = asyncio.Queue()
    ctxt = Context("pva")

    a_monitor = ctxt.monitor(f"{pv_prefix}:A_RBV", a_values.put)

    try:
        value = (await a_values.get()).raw
        assert value["value"] == 0
        assert isinstance(value["value"], int)
        assert value["alarm"]["severity"] == 0
        assert value["alarm"]["message"] == "No alarm"

        await ctxt.put(f"{pv_prefix}:A", 40_001)

        value = (await a_values.get()).raw
        assert value["value"] == 40_001
        assert isinstance(value["value"], int)
        assert value["alarm"]["severity"] == 2
        assert value["alarm"]["message"] == "Above maximum alarm limit: 40000"

        await ctxt.put(f"{pv_prefix}:A", 40_000)

        value = (await a_values.get()).raw
        assert value["value"] == 40_000
        assert isinstance(value["value"], int)
        assert value["alarm"]["severity"] == 0
        assert value["alarm"]["message"] == "No alarm"

        assert a_values.empty()

    finally:
        a_monitor.close()


def make_fastcs(pv_prefix: str, controller: Controller) -> FastCS:
    return FastCS(
        controller, [EpicsPVATransport(pva_ioc=EpicsIOCOptions(pv_prefix=pv_prefix))]
    )


def test_read_signal_set():
    class SomeController(Controller):
        a: AttrRW = AttrRW(Int(max=400_000, max_alarm=40_000))
        b: AttrR = AttrR(Float(min=-1, min_alarm=-0.5, prec=2))

    controller = SomeController()
    pv_prefix = str(uuid4())
    fastcs = make_fastcs(pv_prefix, controller)

    ctxt = ThreadContext("pva")

    async def _wait_and_set_attr_r():
        await asyncio.sleep(0.05)
        await controller.a.update(40_000)
        await controller.b.update(-0.99)
        await asyncio.sleep(0.05)
        await controller.a.update(-100)
        await controller.b.update(-0.99)
        await controller.b.update(-0.9111111)

    a_values, b_values = [], []
    a_monitor = ctxt.monitor(f"{pv_prefix}:A_RBV", a_values.append)
    b_monitor = ctxt.monitor(f"{pv_prefix}:B", b_values.append)
    serve = asyncio.ensure_future(fastcs.serve(interactive=False))
    wait_and_set_attr_r = asyncio.ensure_future(_wait_and_set_attr_r())
    try:
        asyncio.get_event_loop().run_until_complete(
            asyncio.wait_for(asyncio.gather(serve, wait_and_set_attr_r), timeout=0.2)
        )
    except TimeoutError:
        ...
    finally:
        a_monitor.close()
        b_monitor.close()
        serve.cancel()
        wait_and_set_attr_r.cancel()
        assert a_values == [0, 40_000, -100]
        assert b_values == [0.0, -0.99, -0.99, -0.91]  # Last is -0.91 because of prec


def test_pvi_grouping():
    class ChildChildController(Controller):
        attr_e: AttrRW = AttrRW(Int())
        attr_f: AttrR = AttrR(String())

    class ChildController(Controller):
        attr_c: AttrW = AttrW(Bool(), description="Some bool")
        attr_d: AttrW = AttrW(String())

    class SomeController(Controller):
        description = "some controller"
        attr_1: AttrRW = AttrRW(Int(max=400_000, max_alarm=40_000))
        attr_1: AttrRW = AttrRW(Float(min=-1, min_alarm=-0.5, prec=2))
        another_attr_0: AttrRW = AttrRW(Int())
        another_attr_1000: AttrRW = AttrRW(Int())
        a_third_attr: AttrW = AttrW(Int())

    controller = SomeController()

    sub_controller = ChildController()
    controller.child0 = sub_controller
    sub_controller.child_child = ChildChildController()

    sub_controller = ChildController()
    controller.child1 = sub_controller
    sub_controller.child_child = ChildChildController()

    sub_controller = ChildController()
    controller.child2 = sub_controller
    sub_controller.child_child = ChildChildController()

    sub_controller = ChildController()
    controller.another_child = sub_controller
    sub_controller.child_child = ChildChildController()

    sub_controller = ChildController()
    controller.additional_child = sub_controller
    sub_controller.child_child = ChildChildController()

    pv_prefix = str(uuid4())
    fastcs = make_fastcs(pv_prefix, controller)

    ctxt = ThreadContext("pva")

    controller_pvi, child_controller_pvi, child_child_controller_pvi = [], [], []
    controller_monitor = ctxt.monitor(f"{pv_prefix}:PVI", controller_pvi.append)
    child_controller_monitor = ctxt.monitor(
        f"{pv_prefix}:Child0:PVI", child_controller_pvi.append
    )
    child_child_controller_monitor = ctxt.monitor(
        f"{pv_prefix}:Child0:ChildChild:PVI", child_child_controller_pvi.append
    )
    serve = asyncio.ensure_future(fastcs.serve(interactive=False))

    try:
        asyncio.get_event_loop().run_until_complete(
            asyncio.wait_for(serve, timeout=0.2)
        )
    except TimeoutError:
        ...
    finally:
        controller_monitor.close()
        child_controller_monitor.close()
        child_child_controller_monitor.close()
        serve.cancel()

        assert len(controller_pvi) == 1
        assert controller_pvi[0].todict() == {
            "alarm": {"message": "", "severity": 0, "status": 0},
            "display": {"description": "some controller"},
            "timeStamp": {
                "nanoseconds": ANY,
                "secondsPastEpoch": ANY,
                "userTag": 0,
            },
            "value": {
                "additional_child": {"d": f"{pv_prefix}:AdditionalChild:PVI"},
                "another_child": {"d": f"{pv_prefix}:AnotherChild:PVI"},
                "another_attr0": {"rw": f"{pv_prefix}:AnotherAttr0"},
                "another_attr1000": {"rw": f"{pv_prefix}:AnotherAttr1000"},
                "a_third_attr": {"w": f"{pv_prefix}:AThirdAttr"},
                "attr1": {"rw": f"{pv_prefix}:Attr1"},
                "child": {
                    "d": {
                        "v0": f"{pv_prefix}:Child0:PVI",
                        "v1": f"{pv_prefix}:Child1:PVI",
                        "v2": f"{pv_prefix}:Child2:PVI",
                    }
                },
            },
        }
        assert len(child_controller_pvi) == 1
        assert child_controller_pvi[0].todict() == {
            "alarm": {"message": "", "severity": 0, "status": 0},
            "display": {"description": ""},
            "timeStamp": {
                "nanoseconds": ANY,
                "secondsPastEpoch": ANY,
                "userTag": 0,
            },
            "value": {
                "attr_c": {"w": f"{pv_prefix}:Child0:AttrC"},
                "attr_d": {
                    "w": f"{pv_prefix}:Child0:AttrD",
                },
                "child_child": {"d": f"{pv_prefix}:Child0:ChildChild:PVI"},
            },
        }
        assert len(child_child_controller_pvi) == 1
        assert child_child_controller_pvi[0].todict() == {
            "alarm": {"message": "", "severity": 0, "status": 0},
            "display": {"description": ""},
            "timeStamp": {
                "nanoseconds": ANY,
                "secondsPastEpoch": ANY,
                "userTag": 0,
            },
            "value": {
                "attr_e": {"rw": f"{pv_prefix}:Child0:ChildChild:AttrE"},
                "attr_f": {"r": f"{pv_prefix}:Child0:ChildChild:AttrF"},
            },
        }


def test_more_exotic_dataypes():
    table_columns: list[tuple[str, DTypeLike]] = [
        ("A", "i"),
        ("B", "i"),
        ("C", "?"),
        ("D", "f"),
        ("E", "h"),
    ]

    class AnEnum(enum.Enum):
        A = 1
        B = 0
        C = 3

    class SomeController(Controller):
        some_waveform: AttrRW = AttrRW(Waveform(np.int64, shape=(10, 10)))
        some_table: AttrRW = AttrRW(Table(table_columns))
        some_enum: AttrRW = AttrRW(Enum(AnEnum))

    controller = SomeController()
    pv_prefix = str(uuid4())
    fastcs = make_fastcs(pv_prefix, controller)

    ctxt = ThreadContext("pva", nt=False)

    initial_waveform_value = np.zeros((10, 10), dtype=np.int64)
    initial_table_value = np.array([], dtype=table_columns)
    initial_enum_value = AnEnum.A

    server_set_waveform_value = np.copy(initial_waveform_value)
    server_set_waveform_value[0] = np.arange(10)
    server_set_table_value = np.array([(1, 2, False, 3.14, 1)], dtype=table_columns)
    server_set_enum_value = AnEnum.B

    client_put_waveform_value = np.copy(server_set_waveform_value)
    client_put_waveform_value[1] = np.arange(10)
    client_put_table_value = NTTable(columns=table_columns).wrap(
        [
            {"A": 1, "B": 2, "C": False, "D": 3.14, "E": 1},
            {"A": 5, "B": 2, "C": True, "D": 6.28, "E": 2},
        ]
    )
    client_put_enum_value = "C"

    async def _wait_and_set_attrs():
        await asyncio.sleep(0.1)
        # This demonstrates an update from hardware,
        # resulting in only a change in the read back.
        await asyncio.gather(
            controller.some_waveform.update(server_set_waveform_value),
            controller.some_table.update(server_set_table_value),
            controller.some_enum.update(server_set_enum_value),
        )

    async def _wait_and_put_pvs():
        await asyncio.sleep(0.3)
        ctxt = Context("pva")
        # This demonstrates a client put,
        # resulting in a change in the demand and read back.
        await asyncio.gather(
            ctxt.put(f"{pv_prefix}:SomeWaveform", client_put_waveform_value),
            ctxt.put(f"{pv_prefix}:SomeTable", client_put_table_value),
            ctxt.put(f"{pv_prefix}:SomeEnum", client_put_enum_value),
        )

    waveform_values, table_values, enum_values = [], [], []

    # Monitoring read backs to capture both client and server sets.
    waveform_monitor = ctxt.monitor(
        f"{pv_prefix}:SomeWaveform_RBV", waveform_values.append
    )
    table_monitor = ctxt.monitor(f"{pv_prefix}:SomeTable_RBV", table_values.append)
    enum_monitor = ctxt.monitor(
        f"{pv_prefix}:SomeEnum_RBV",
        enum_values.append,
    )

    serve = asyncio.ensure_future(fastcs.serve(interactive=False))
    wait_and_set_attrs = asyncio.ensure_future(_wait_and_set_attrs())
    wait_and_put_pvs = asyncio.ensure_future(_wait_and_put_pvs())
    try:
        asyncio.get_event_loop().run_until_complete(
            asyncio.wait_for(
                asyncio.gather(serve, wait_and_set_attrs, wait_and_put_pvs),
                timeout=0.6,
            )
        )
    except TimeoutError:
        ...
    finally:
        waveform_monitor.close()
        table_monitor.close()
        enum_monitor.close()
        serve.cancel()
        wait_and_set_attrs.cancel()
        wait_and_put_pvs.cancel()

        expected_waveform_gets = [
            initial_waveform_value,
            server_set_waveform_value,
            client_put_waveform_value,
        ]

        for expected_waveform, actual_waveform in zip(
            expected_waveform_gets, waveform_values, strict=True
        ):
            np.testing.assert_array_equal(
                expected_waveform, actual_waveform.todict()["value"].reshape(10, 10)
            )

        expected_table_gets = [
            NTTable(columns=table_columns).wrap(initial_table_value),
            NTTable(columns=table_columns).wrap(server_set_table_value),
            client_put_table_value,
        ]
        for expected_table, actual_table in zip(
            expected_table_gets, table_values, strict=True
        ):
            expected_table = expected_table.todict()["value"]
            actual_table = actual_table.todict()["value"]
            for expected_column, actual_column in zip(
                expected_table.values(), actual_table.values(), strict=True
            ):
                if isinstance(expected_column, np.ndarray):
                    np.testing.assert_array_equal(expected_column, actual_column)
                else:
                    assert expected_column == actual_column and actual_column is None

        expected_enum_gets = [
            initial_enum_value,
            server_set_enum_value,
            AnEnum.C,
        ]

        for expected_enum, actual_enum in zip(
            expected_enum_gets, enum_values, strict=True
        ):
            assert (
                expected_enum
                == controller.some_enum.datatype.members[  # type: ignore
                    actual_enum.todict()["value"]["index"]
                ]
            )


def test_command_method_put_twice(caplog):
    class SomeController(Controller):
        command_runs_for_a_while_times = []
        command_spawns_a_task_times = []
        command_task_times = []

        @command()
        async def command_runs_for_a_while(self):
            start_time = datetime.now()
            await asyncio.sleep(0.1)
            self.command_runs_for_a_while_times.append((start_time, datetime.now()))

        @command()
        async def command_spawns_a_task(self):
            start_time = datetime.now()

            async def some_task():
                task_start_time = datetime.now()
                await asyncio.sleep(0.1)
                self.command_task_times.append((task_start_time, datetime.now()))

            self.command_spawns_a_task_times.append((start_time, datetime.now()))

            asyncio.create_task(some_task())

    controller = SomeController()
    pv_prefix = str(uuid4())
    fastcs = make_fastcs(pv_prefix, controller)
    expected_error_string = (
        "RuntimeError: Received request to run command but it is "
        "already in progress. Maybe the command should spawn an asyncio task?"
    )

    async def put_pvs():
        await asyncio.sleep(0.1)
        ctxt = Context("pva")
        await asyncio.gather(
            ctxt.put(f"{pv_prefix}:CommandSpawnsATask", True),
            ctxt.put(f"{pv_prefix}:CommandSpawnsATask", True),
        )
        assert expected_error_string not in caplog.text
        await asyncio.gather(
            ctxt.put(f"{pv_prefix}:CommandRunsForAWhile", True),
            ctxt.put(f"{pv_prefix}:CommandRunsForAWhile", True),
        )
        assert expected_error_string in caplog.text

    serve = asyncio.ensure_future(fastcs.serve(interactive=False))
    try:
        asyncio.get_event_loop().run_until_complete(
            asyncio.wait_for(
                asyncio.gather(serve, put_pvs()),
                timeout=3,
            )
        )
    except TimeoutError:
        ...
    serve.cancel()

    assert (
        len(controller.command_task_times)
        == len(controller.command_spawns_a_task_times)
        == 2
    )
    for (task_start_time, task_end_time), (
        task_spawn_start_time,
        task_spawn_end_time,
    ) in zip(
        controller.command_task_times,
        controller.command_spawns_a_task_times,
        strict=True,
    ):
        assert (
            pytest.approx(
                (task_spawn_end_time - task_spawn_start_time).total_seconds(), abs=0.05
            )
            == 0
        )
        assert (
            pytest.approx((task_end_time - task_start_time).total_seconds(), abs=0.05)
            == 0.1
        )

    assert len(controller.command_runs_for_a_while_times) == 1
    coro_start_time, coro_end_time = controller.command_runs_for_a_while_times[0]
    assert (
        pytest.approx((coro_end_time - coro_start_time).total_seconds(), abs=0.05)
        == 0.1
    )


def test_block_flag_waits_for_callback_completion():
    class SomeController(Controller):
        @command()
        async def command_runs_for_a_while(self):
            await asyncio.sleep(0.2)

    controller = SomeController()
    pv_prefix = str(uuid4())
    fastcs = make_fastcs(pv_prefix, controller)
    command_runs_for_a_while_times = []

    async def put_pvs():
        ctxt = Context("pva")
        for block in [True, False]:
            start_time = datetime.now()
            await ctxt.put(
                f"{pv_prefix}:CommandRunsForAWhile",
                True,
                wait=block,
            )
            command_runs_for_a_while_times.append((start_time, datetime.now()))

    serve = asyncio.ensure_future(fastcs.serve(interactive=False))
    try:
        asyncio.get_event_loop().run_until_complete(
            asyncio.wait_for(
                asyncio.gather(serve, put_pvs()),
                timeout=0.5,
            )
        )
    except TimeoutError:
        ...
    serve.cancel()

    assert len(command_runs_for_a_while_times) == 2

    for put_call, expected_duration in enumerate([0.2, 0]):
        start, end = command_runs_for_a_while_times[put_call]
        assert (
            pytest.approx((end - start).total_seconds(), abs=0.1) == expected_duration
        )



import contextlib
import multiprocessing
import os

import pytest
import requests
import tango
from p4p.client.thread import Context

FASTCS_BENCHMARKING = os.getenv("FASTCS_BENCHMARKING") == "true"
GET_ENDPOINT = "http://localhost:8090/read-int"
PUT_ENDPOINT = "http://localhost:8090/write-bool"
GET_PV = "BENCHMARK-DEVICE:ReadInt"
PUT_PV = "BENCHMARK-DEVICE:WriteBool"
TANGO_DEVICE = "MY/BENCHMARK/DEVICE"
READ_ATTR = "ReadInt"
WRITE_ATTR = "WriteBool"
REST_CLIENTS = 9
os.environ["TANGO_HOST"] = "localhost:10000"


def rest_get():
    requests.get(GET_ENDPOINT)


def rest_put():
    requests.put(PUT_ENDPOINT, json={"value": "false"})


ctx = Context("pva")


def epics_get():
    ctx.get(GET_PV)


def epics_put():
    ctx.put(PUT_PV, 0)


def bg_get(event, url):
    """Workload for a rest subprocess."""
    while not event.is_set():
        requests.get(url)


def bg_pass(event, url):
    """Workload for a rest subprocess."""
    while not event.is_set():
        pass


@contextlib.contextmanager
def background_traffic(rest_target):
    """Context manager to manage background clients"""
    stop_event = multiprocessing.Event()
    processes = [
        multiprocessing.Process(
            target=bg_get if rest_target else bg_pass,
            args=(stop_event, rest_target),
        )
        for _ in range(REST_CLIENTS)
    ]
    for process in processes:
        process.start()

    try:
        yield processes
    finally:
        # Signal the processes to stop
        stop_event.set()
        # Wait for the processes to finish
        for process in processes:
            process.join()


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-rest")
def test_rest_get(benchmark, test_controller):
    benchmark(rest_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-rest")
def test_rest_get_loaded_request(benchmark, test_controller):
    with background_traffic(GET_ENDPOINT):
        benchmark(rest_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-rest")
def test_rest_get_loaded_baseline(benchmark, test_controller):
    with background_traffic(None):
        benchmark(rest_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-rest")
def test_rest_put(benchmark, test_controller):
    benchmark(rest_put)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-epics")
def test_epics_get(benchmark, test_controller):
    benchmark(epics_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-epics")
def test_epics_get_loaded_request(benchmark, test_controller):
    with background_traffic(GET_ENDPOINT):
        benchmark(epics_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-epics")
def test_epics_get_loaded_baseline(benchmark, test_controller):
    with background_traffic(None):
        benchmark(epics_get)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-epics")
def test_epics_put(benchmark, test_controller):
    benchmark(epics_put)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-tango")
def test_tango_get(benchmark, test_controller):
    device = tango.DeviceProxy(TANGO_DEVICE)

    def to_do():
        device.read_attribute(READ_ATTR)

    benchmark(to_do)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-tango")
def test_tango_get_loaded_request(benchmark, test_controller):
    device = tango.DeviceProxy(TANGO_DEVICE)

    def to_do():
        device.read_attribute(READ_ATTR)

    with background_traffic(GET_ENDPOINT):
        benchmark(to_do)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-tango")
def test_tango_get_loaded_baseline(benchmark, test_controller):
    device = tango.DeviceProxy(TANGO_DEVICE)

    def to_do():
        device.read_attribute(READ_ATTR)

    with background_traffic(None):
        benchmark(to_do)


@pytest.mark.skipif(not FASTCS_BENCHMARKING, reason="export FASTCS_BENCHMARKING=true")
@pytest.mark.benchmark(group="test-tango")
def test_tango_put(benchmark, test_controller):
    device = tango.DeviceProxy(TANGO_DEVICE)

    def to_do():
        device.write_attribute(WRITE_ATTR, 0)

    benchmark(to_do)



from collections import defaultdict
from typing import Any

from fastcs.logging import logger


class Tracer:
    """A mixin or standalone class for conditionally logging trace events.

    This can be used for verbose logging that is disabled by default and enabled on a
    per-instance basis, with filtering based on specific key-value pairs on the event.

    Any instance of this class can enable tracing independently. Some key classes
    inherit this class, such as ``Attributes``, and some modules have their own
    ``Tracer``, such as ``fastcs.launch``. When enabled, any event logged from the
    object, or from another ``Tracer`` that uses the object as the ``topic``, will be
    logged.

    Note: The global logger level must be set to ``TRACE`` for the messages to be logged

    Example usage:
    .. code-block:: python

        controller.ramp_rate.enable_tracing()
        controller.ramp_rate.disable_tracing()
        controller.connection.enable_tracing()
        controller.connection.add_tracing_filter("query", "V?")
        controller.connection.remove_tracing_filter("query", "V?")
        controller.connection.disable_tracing()

    :param name: The name of the logger. Attached to log messages as ``logger_name``.

    """

    def __init__(self, name: str | None = None):
        self.__tracing_enabled: bool = False
        self.__tracing_filters: dict[str, list[Any]] = defaultdict(list)
        self.__logger_name = name if name is not None else self.__class__.__name__

    def log_event(self, event: str, topic: "Tracer | None" = None, *args, **kwargs):
        """Log an event only if tracing is enabled and the filter matches

        :param event: A message describing the event
        :param topic: Another `Tracer` related to this event to enable it to be logged
        :param args: Positional arguments for underlying logger
        :param kwargs: Keyword arguments for underlying logger

        """
        if self.__tracing_enabled or (topic is not None and topic.__tracing_enabled):  # noqa: SLF001
            if self.__tracing_filters:
                for kwarg, values in self.__tracing_filters.items():
                    if kwarg in kwargs and any(kwargs[kwarg] == v for v in values):
                        break
                else:
                    return

            logger.trace(event, *args, logger_name=self.__logger_name, **kwargs)

    def enable_tracing(self):
        """Enable trace logging for this object"""
        self.__tracing_enabled = True

    def disable_tracing(self):
        """Disable trace logging for this object"""
        self.__tracing_enabled = False

    def add_tracing_filter(self, key: str, value: Any):
        """Add a filter for trace log messages from this object

        To reduce trace messages further, a filter can be applied such that events must
        have a key with a specific value for it to be logged.

        :param key: A new or existing key to filter on
        :param value: An allowed value for the event to be logged

        """
        self.__tracing_filters[key].append(value)

    def remove_tracing_filter(self, key: str, value: Any):
        """Remove a specific key-value pair from the filter

        :param key: An existing filter key
        :param value: The allowed value to remove

        """
        if (
            key not in self.__tracing_filters
            or value not in self.__tracing_filters[key]
        ):
            return

        self.__tracing_filters[key].remove(value)
        if not self.__tracing_filters[key]:
            self.__tracing_filters.pop(key)

    def clear_tracing_filters(self):
        """Clear all filters and allow all messages to be logged (if enabled)"""
        self.__tracing_filters.clear()



from collections.abc import Callable

from .cs_methods import (
    Controller_T,
    UnboundCommand,
    UnboundCommandCallback,
    UnboundScan,
    UnboundScanCallback,
)
from .exceptions import FastCSError


def scan(
    period: float,
) -> Callable[[UnboundScanCallback[Controller_T]], UnboundScan[Controller_T]]:
    """Sets up a scan over the wrapped method."""

    if period <= 0:
        raise FastCSError("Scan method must have a positive scan period")

    def wrapper(fn: UnboundScanCallback[Controller_T]) -> UnboundScan[Controller_T]:
        return UnboundScan(fn, period)

    return wrapper


def command(
    *, group: str | None = None
) -> Callable[[UnboundCommandCallback[Controller_T]], UnboundCommand[Controller_T]]:
    """Decorator to tag a `Controller` method to be turned into a `Command`.

    Args:
        group: Group to display this command under in the transport layer

    """

    def wrapper(
        fn: UnboundCommandCallback[Controller_T],
    ) -> UnboundCommand[Controller_T]:
        return UnboundCommand(fn, group=group)

    return wrapper



from fastcs.controller_api import ControllerAPI

from .options import EpicsDocsOptions


class EpicsDocs:
    """For creating docs in the EPICS transports."""

    def __init__(self, controller_apis: ControllerAPI) -> None:
        self._controller_apis = controller_apis

    def create_docs(self, options: EpicsDocsOptions | None = None) -> None:
        if options is None:
            options = EpicsDocsOptions()



import asyncio
from dataclasses import dataclass, field

from fastcs.controller_api import ControllerAPI
from fastcs.logging import logger as _fastcs_logger
from fastcs.transport.epics.docs import EpicsDocs
from fastcs.transport.epics.gui import PvaEpicsGUI
from fastcs.transport.epics.options import (
    EpicsDocsOptions,
    EpicsGUIOptions,
    EpicsIOCOptions,
)
from fastcs.transport.transport import Transport

from .ioc import P4PIOC

logger = _fastcs_logger.bind(logger_name=__name__)


@dataclass
class EpicsPVATransport(Transport):
    """PV access transport."""

    pva_ioc: EpicsIOCOptions = field(default_factory=EpicsIOCOptions)
    docs: EpicsDocsOptions | None = None
    gui: EpicsGUIOptions | None = None

    def connect(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self._controller_api = controller_api
        self._pv_prefix = self.pva_ioc.pv_prefix
        self._ioc = P4PIOC(self.pva_ioc.pv_prefix, controller_api)

        if self.docs is not None:
            EpicsDocs(self._controller_api).create_docs(self.docs)

        if self.gui is not None:
            PvaEpicsGUI(self._controller_api, self._pv_prefix).create_gui(self.gui)

    async def serve(self) -> None:
        logger.info("Running IOC", pv_prefix=self._pv_prefix)
        await self._ioc.run()

    def __repr__(self):
        return f"EpicsPVATransport({self._pv_prefix})"



import asyncio
import copy
import json
from typing import Any

import pytest
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture
from tests.assertable_controller import (
    AssertableControllerAPI,
    MyTestAttributeIORef,
    MyTestController,
)

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.datatypes import Bool, Float, Int, String
from fastcs.transport.graphql.transport import GraphQLTransport


class GraphQLController(MyTestController):
    read_int = AttrR(Int(), io_ref=MyTestAttributeIORef())
    read_write_int = AttrRW(Int(), io_ref=MyTestAttributeIORef())
    read_write_float = AttrRW(Float())
    read_bool = AttrR(Bool())
    write_bool = AttrW(Bool(), io_ref=MyTestAttributeIORef())
    read_string = AttrRW(String())


@pytest.fixture(scope="class")
def gql_controller_api(class_mocker: MockerFixture):
    return AssertableControllerAPI(GraphQLController(), class_mocker)


def nest_query(path: list[str]) -> str:
    queue = copy.deepcopy(path)
    field = queue.pop(0)

    if queue:
        nesting = nest_query(queue)
        return f"{field} {{ {nesting} }} "
    else:
        return field


def nest_mutation(path: list[str], value: Any) -> str:
    queue = copy.deepcopy(path)
    field = queue.pop(0)

    if queue:
        nesting = nest_query(queue)
        return f"{field} {{ {nesting} }} "
    else:
        return f"{field}(value: {json.dumps(value)})"


def nest_response(path: list[str], value: Any) -> dict:
    queue = copy.deepcopy(path)
    field = queue.pop(0)

    if queue:
        nesting = nest_response(queue, value)
        return {field: nesting}
    else:
        return {field: value}


def create_test_client(gql_controller_api: AssertableControllerAPI) -> TestClient:
    graphql_transport = GraphQLTransport()
    graphql_transport.connect(gql_controller_api, asyncio.AbstractEventLoop())
    return TestClient(graphql_transport._server._app)


class TestGraphQLServer:
    @pytest.fixture(scope="class")
    def test_client(self, gql_controller_api) -> TestClient:
        return create_test_client(gql_controller_api)

    def test_read_int(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        path = ["readInt"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["read_int"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)

    def test_read_write_int(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        path = ["readWriteInt"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["read_write_int"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)

        new = 9
        mutation = f"mutation {{ {nest_mutation(path, new)} }}"
        with gql_controller_api.assert_write_here(["read_write_int"]):
            response = test_client.post("/graphql", json={"query": mutation})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, new)

    def test_read_write_float(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        path = ["readWriteFloat"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["read_write_float"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)

        new = 0.5
        mutation = f"mutation {{ {nest_mutation(path, new)} }}"
        with gql_controller_api.assert_write_here(["read_write_float"]):
            response = test_client.post("/graphql", json={"query": mutation})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, new)

    def test_read_bool(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = False
        path = ["readBool"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["read_bool"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)

    def test_write_bool(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        value = True
        path = ["writeBool"]
        mutation = f"mutation {{ {nest_mutation(path, value)} }}"
        with gql_controller_api.assert_write_here(["write_bool"]):
            response = test_client.post("/graphql", json={"query": mutation})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, value)

    def test_go(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        test_client = create_test_client(gql_controller_api)

        path = ["go"]
        mutation = f"mutation {{ {nest_query(path)} }}"
        with gql_controller_api.assert_execute_here(["go"]):
            response = test_client.post("/graphql", json={"query": mutation})

        assert response.status_code == 200
        assert response.json()["data"] == {path[-1]: True}

    def test_read_child1(
        self, gql_controller_api: AssertableControllerAPI, test_client: TestClient
    ):
        expect = 0
        path = ["SubController01", "readInt"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["SubController01", "read_int"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)

    def test_read_child2(self, gql_controller_api, test_client: TestClient):
        expect = 0
        path = ["SubController02", "readInt"]
        query = f"query {{ {nest_query(path)} }}"
        with gql_controller_api.assert_read_here(["SubController02", "read_int"]):
            response = test_client.post("/graphql", json={"query": query})
        assert response.status_code == 200
        assert response.json()["data"] == nest_response(path, expect)



import asyncio
from dataclasses import dataclass, field
from typing import Any

from softioc import softioc

from fastcs.controller_api import ControllerAPI
from fastcs.logging import logger as _fastcs_logger
from fastcs.transport.epics.ca.ioc import EpicsCAIOC
from fastcs.transport.epics.docs import EpicsDocs
from fastcs.transport.epics.gui import EpicsGUI
from fastcs.transport.epics.options import (
    EpicsDocsOptions,
    EpicsGUIOptions,
    EpicsIOCOptions,
)
from fastcs.transport.transport import Transport

logger = _fastcs_logger.bind(logger_name=__name__)


@dataclass
class EpicsCATransport(Transport):
    """Channel access transport."""

    ca_ioc: EpicsIOCOptions = field(default_factory=EpicsIOCOptions)
    docs: EpicsDocsOptions | None = None
    gui: EpicsGUIOptions | None = None

    def connect(
        self,
        controller_api: ControllerAPI,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self._controller_api = controller_api
        self._loop = loop
        self._pv_prefix = self.ca_ioc.pv_prefix
        self._ioc = EpicsCAIOC(self.ca_ioc.pv_prefix, controller_api, self.ca_ioc)

        if self.docs is not None:
            EpicsDocs(self._controller_api).create_docs(self.docs)

        if self.gui is not None:
            EpicsGUI(self._controller_api, self._pv_prefix).create_gui(self.gui)

    async def serve(self) -> None:
        logger.info("Running IOC", pv_prefix=self._pv_prefix)
        self._ioc.run(self._loop)

    @property
    def context(self) -> dict[str, Any]:
        return {
            command_name: getattr(softioc, command_name)
            for command_name in softioc.command_names
            if command_name != "exit"
        }

    def __repr__(self):
        return f"EpicsCATransport({self._pv_prefix})"



from dataclasses import dataclass, field

from ..options import (
    EpicsDocsOptions,
    EpicsGUIOptions,
    EpicsIOCOptions,
)


@dataclass
class EpicsCAOptions:
    """Options for the EPICS CA transport."""

    docs: EpicsDocsOptions = field(default_factory=EpicsDocsOptions)
    gui: EpicsGUIOptions = field(default_factory=EpicsGUIOptions)
    ca_ioc: EpicsIOCOptions = field(default_factory=EpicsIOCOptions)



import asyncio
import enum

import numpy as np
import pytest
from pytest_mock import MockerFixture
from tango import DeviceProxy, DevState
from tango.test_context import DeviceTestContext
from tests.assertable_controller import (
    AssertableControllerAPI,
    MyTestController,
)

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.datatypes import Bool, Enum, Float, Int, String, Waveform
from fastcs.transport.tango.transport import TangoTransport


async def patch_run_threadsafe_blocking(coro, loop):
    await coro


@pytest.fixture(scope="module")
def mock_run_threadsafe_blocking(module_mocker: MockerFixture):
    m = module_mocker.patch(
        "fastcs.transport.tango.dsr._run_threadsafe_blocking",
        patch_run_threadsafe_blocking,
    )
    yield m


class TangoController(MyTestController):
    read_int = AttrR(Int())
    read_write_int = AttrRW(Int())
    read_write_float = AttrRW(Float())
    read_bool = AttrR(Bool())
    write_bool = AttrW(Bool())
    read_string = AttrRW(String())
    enum = AttrRW(Enum(enum.IntEnum("Enum", {"RED": 0, "GREEN": 1, "BLUE": 2})))
    one_d_waveform = AttrRW(Waveform(np.int32, (10,)))
    two_d_waveform = AttrRW(Waveform(np.int32, (10, 10)))


@pytest.fixture(scope="class")
def tango_controller_api(class_mocker: MockerFixture) -> AssertableControllerAPI:
    return AssertableControllerAPI(TangoController(), class_mocker)


def create_test_context(tango_controller_api: AssertableControllerAPI):
    tango_transport = TangoTransport()
    tango_transport.connect(
        tango_controller_api,
        # This is passed to enable instantiating the transport, but tests must avoid
        # using via patching of functions. It will raise NotImplementedError if used.
        asyncio.AbstractEventLoop(),
    )
    device = tango_transport._dsr._device
    # https://tango-controls.readthedocs.io/projects/pytango/en/v9.5.1/testing/test_context.html
    with DeviceTestContext(device, debug=0) as proxy:
        yield proxy


class TestTangoDevice:
    @pytest.fixture(scope="class")
    def tango_context(
        self,
        mock_run_threadsafe_blocking,
        tango_controller_api: AssertableControllerAPI,
    ):
        yield from create_test_context(tango_controller_api)

    def test_list_attributes(self, tango_context: DeviceProxy):
        assert list(tango_context.get_attribute_list()) == [
            "Enum",
            "OneDWaveform",
            "ReadBool",
            "ReadInt",
            "ReadString",
            "ReadWriteFloat",
            "ReadWriteInt",
            "TwoDWaveform",
            "WriteBool",
            "SubController01_ReadInt",
            "SubController02_ReadInt",
            "State",
            "Status",
        ]

    def test_list_commands(self, tango_context):
        assert list(tango_context.get_command_list()) == [
            "Go",
            "Init",
            "State",
            "Status",
        ]

    def test_state(self, tango_context):
        assert tango_context.command_inout("State") == DevState.ON

    def test_status(self, tango_context):
        expect = "The device is in ON state."
        assert tango_context.command_inout("Status") == expect

    def test_read_int(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = 0
        with tango_controller_api.assert_read_here(["read_int"]):
            result = tango_context.read_attribute("ReadInt").value
        assert result == expect

    def test_read_write_int(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = 0
        with tango_controller_api.assert_read_here(["read_write_int"]):
            result = tango_context.read_attribute("ReadWriteInt").value
        assert result == expect
        new = 9
        with tango_controller_api.assert_write_here(["read_write_int"]):
            tango_context.write_attribute("ReadWriteInt", new)
        assert tango_context.read_attribute("ReadWriteInt").value == new

    def test_read_write_float(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = 0.0
        with tango_controller_api.assert_read_here(["read_write_float"]):
            result = tango_context.read_attribute("ReadWriteFloat").value
        assert result == expect
        new = 0.5
        with tango_controller_api.assert_write_here(["read_write_float"]):
            tango_context.write_attribute("ReadWriteFloat", new)
        assert tango_context.read_attribute("ReadWriteFloat").value == new

    def test_read_bool(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = False
        with tango_controller_api.assert_read_here(["read_bool"]):
            result = tango_context.read_attribute("ReadBool").value
        assert result == expect

    def test_write_bool(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        with tango_controller_api.assert_write_here(["write_bool"]):
            tango_context.write_attribute("WriteBool", True)

    def test_enum(self, tango_controller_api: AssertableControllerAPI, tango_context):
        enum_attr = tango_controller_api.attributes["enum"]
        assert isinstance(enum_attr, AttrRW)
        enum_cls = enum_attr.datatype.dtype
        assert isinstance(enum_attr.get(), enum_cls)
        assert enum_attr.get() == enum_cls(0)
        expect = 0
        with tango_controller_api.assert_read_here(["enum"]):
            result = tango_context.read_attribute("Enum").value
        assert result == expect
        new = 1
        with tango_controller_api.assert_write_here(["enum"]):
            tango_context.write_attribute("Enum", new)
        assert tango_context.read_attribute("Enum").value == new
        assert isinstance(enum_attr.get(), enum_cls)
        assert enum_attr.get() == enum_cls(1)

    def test_1d_waveform(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = np.zeros((10,), dtype=np.int32)
        with tango_controller_api.assert_read_here(["one_d_waveform"]):
            result = tango_context.read_attribute("OneDWaveform").value
        assert np.array_equal(result, expect)
        new = np.array([1, 2, 3], dtype=np.int32)
        with tango_controller_api.assert_write_here(["one_d_waveform"]):
            tango_context.write_attribute("OneDWaveform", new)
        assert np.array_equal(tango_context.read_attribute("OneDWaveform").value, new)

    def test_2d_waveform(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = np.zeros((10, 10), dtype=np.int32)
        with tango_controller_api.assert_read_here(["two_d_waveform"]):
            result = tango_context.read_attribute("TwoDWaveform").value
        assert np.array_equal(result, expect)
        new = np.array([[1, 2, 3]], dtype=np.int32)
        with tango_controller_api.assert_write_here(["two_d_waveform"]):
            tango_context.write_attribute("TwoDWaveform", new)
        assert np.array_equal(tango_context.read_attribute("TwoDWaveform").value, new)

    def test_read_child1(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = 0
        with tango_controller_api.assert_read_here(["SubController01", "read_int"]):
            result = tango_context.read_attribute("SubController01_ReadInt").value
        assert result == expect

    def test_read_child2(
        self, tango_controller_api: AssertableControllerAPI, tango_context
    ):
        expect = 0
        with tango_controller_api.assert_read_here(["SubController02", "read_int"]):
            result = tango_context.read_attribute("SubController02_ReadInt").value
        assert result == expect

    def test_go(self, tango_controller_api: AssertableControllerAPI, tango_context):
        with tango_controller_api.assert_execute_here(["go"]):
            tango_context.command_inout("Go")



import re
from typing import _GenericAlias, get_args, get_origin, get_type_hints  # type: ignore

import numpy as np

from fastcs.attributes import Attribute
from fastcs.controller import BaseController
from fastcs.datatypes import Bool, DataType, Float, Int, String


def snake_to_pascal(name: str) -> str:
    """Converts string from snake case to Pascal case.
    If string is not a valid snake case it will be returned unchanged
    """
    if re.fullmatch(r"[a-z][a-z0-9]*(?:_[a-z0-9]+)*", name):
        name = re.sub(r"(?:^|_)([a-z0-9])", lambda match: match.group(1).upper(), name)
    return name


def numpy_to_fastcs_datatype(np_type) -> DataType:
    """Converts numpy types to fastcs types for widget creation.
    Only types important for widget creation are explicitly converted
    """
    if np.issubdtype(np_type, np.integer):
        return Int()
    elif np.issubdtype(np_type, np.floating):
        return Float()
    elif np.issubdtype(np_type, np.bool_):
        return Bool()
    else:
        return String()


def validate_hinted_attributes(controller: BaseController):
    """Validate ``Attribute`` type-hints  match dynamically intropected instances

    For each type-hinted attribute, validate that a corresponding instance exists in the
    controller with the correct access mode and datatype.
    """

    hints = get_type_hints(type(controller))
    alias_hints = {k: v for k, v in hints.items() if isinstance(v, _GenericAlias)}
    for name, hint in alias_hints.items():
        attr_class = get_origin(hint)
        if not issubclass(attr_class, Attribute):
            continue

        attr = getattr(controller, name, None)
        if attr is None:
            raise RuntimeError(
                f"Controller `{controller.__class__.__name__}` failed to introspect "
                f"hinted attribute `{name}` during initialisation"
            )

        if type(attr) is not attr_class:
            raise RuntimeError(
                f"Controller '{controller.__class__.__name__}' introspection of hinted "
                f"attribute '{name}' does not match defined access mode. "
                f"Expected '{attr_class.__name__}', got '{type(attr).__name__}'."
            )

        attr_dtype = get_args(hint)[0]
        if attr.datatype.dtype != attr_dtype:
            raise RuntimeError(
                f"Controller '{controller.__class__.__name__}' introspection of hinted "
                f"attribute '{name}' does not match defined datatype. "
                f"Expected '{attr_dtype.__name__}', "
                f"got '{attr.datatype.dtype.__name__}'."
            )



import enum

import numpy as np
import pytest
from pvi.device import SignalR
from pydantic import ValidationError

from fastcs.attributes import AttrR, AttrRW
from fastcs.controller import Controller
from fastcs.datatypes import Bool, Enum, Float, Int, String
from fastcs.util import (
    numpy_to_fastcs_datatype,
    snake_to_pascal,
    validate_hinted_attributes,
)


def test_snake_to_pascal():
    name1 = "name_in_snake_case"
    name2 = "name-not-in-snake-case"
    name3 = "name_with-different_separators"
    name4 = "name_with_numbers_1_2_3"
    name5 = "numbers_1_2_3_in_the_middle"
    name6 = "1_2_3_starting_with_numbers"
    name7 = "name1_with2_a3_number4"
    name8 = "name_in_lower_case"
    name9 = "NameAlreadyInPascalCase"
    name10 = "Name_With_%_Invalid_&_Symbols_£_"
    name11 = "a_b_c_d"
    name12 = "test"
    assert snake_to_pascal(name1) == "NameInSnakeCase"
    assert snake_to_pascal(name2) == "name-not-in-snake-case"
    assert snake_to_pascal(name3) == "name_with-different_separators"
    assert snake_to_pascal(name4) == "NameWithNumbers123"
    assert snake_to_pascal(name5) == "Numbers123InTheMiddle"
    assert snake_to_pascal(name6) == "1_2_3_starting_with_numbers"
    assert snake_to_pascal(name7) == "Name1With2A3Number4"
    assert snake_to_pascal(name8) == "NameInLowerCase"
    assert snake_to_pascal(name9) == "NameAlreadyInPascalCase"
    assert snake_to_pascal(name10) == "Name_With_%_Invalid_&_Symbols_£_"
    assert snake_to_pascal(name11) == "ABCD"
    assert snake_to_pascal(name12) == "Test"


def test_pvi_validation_error():
    name = snake_to_pascal("Name-With_%_Invalid-&-Symbols_£_")
    with pytest.raises(ValidationError):
        SignalR(name=name, read_pv="test")


@pytest.mark.parametrize(
    "numpy_type, fastcs_datatype",
    [
        (np.float16, Float()),
        (np.float32, Float()),
        (np.int16, Int()),
        (np.int32, Int()),
        (np.bool, Bool()),
        (np.dtype("S1000"), String()),
        (np.dtype("U25"), String()),
        (np.dtype(">i4"), Int()),
        (np.dtype("d"), Float()),
    ],
)
def test_numpy_to_fastcs_datatype(numpy_type, fastcs_datatype):
    assert fastcs_datatype == numpy_to_fastcs_datatype(numpy_type)


@pytest.mark.asyncio
async def test_hinted_attributes_verified():
    class ControllerWithWrongType(Controller):
        hinted_wrong_type: AttrR[int]

        async def initialise(self):
            self.hinted_wrong_type = AttrR(Float())  # type: ignore

    controller = ControllerWithWrongType()
    await controller.initialise()
    with pytest.raises(RuntimeError) as excinfo:
        validate_hinted_attributes(controller)

    assert str(excinfo.value) == (
        "Controller 'ControllerWithWrongType' introspection of hinted attribute "
        "'hinted_wrong_type' does not match defined datatype. "
        "Expected 'int', got 'float'."
    )

    class ControllerWithMissingAttr(Controller):
        hinted_int_missing: AttrR[int]

    controller = ControllerWithMissingAttr()
    await controller.initialise()
    with pytest.raises(RuntimeError) as excinfo:
        validate_hinted_attributes(controller)

    assert str(excinfo.value) == (
        "Controller `ControllerWithMissingAttr` failed to introspect hinted attribute "
        "`hinted_int_missing` during initialisation"
    )

    class ControllerAttrWrongAccessMode(Controller):
        hinted: AttrR[int]

        async def initialise(self):
            self.hinted = AttrRW(Int())
            self.attributes["hinted"] = self.hinted

    controller = ControllerAttrWrongAccessMode()
    await controller.initialise()
    with pytest.raises(RuntimeError) as excinfo:
        validate_hinted_attributes(controller)

    assert str(excinfo.value) == (
        "Controller 'ControllerAttrWrongAccessMode' introspection of hinted attribute "
        "'hinted' does not match defined access mode. Expected 'AttrR', got 'AttrRW'."
    )

    class MyEnum(enum.Enum):
        A = 0
        B = 1

    class MyEnum2(enum.Enum):
        A = 2
        B = 3

    class ControllerWrongEnumClass(Controller):
        hinted_enum: AttrRW[MyEnum] = AttrRW(Enum(MyEnum2))

    controller = ControllerWrongEnumClass()
    await controller.initialise()
    with pytest.raises(RuntimeError) as excinfo:
        validate_hinted_attributes(controller)

    assert str(excinfo.value) == (
        "Controller 'ControllerWrongEnumClass' introspection of hinted attribute "
        "'hinted_enum' does not match defined datatype. "
        "Expected 'MyEnum', got 'MyEnum2'."
    )



import asyncio
from dataclasses import dataclass

import pytest

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import ONCE, AttrR, AttrRW
from fastcs.control_system import FastCS, build_controller_api
from fastcs.controller import Controller
from fastcs.cs_methods import Command
from fastcs.datatypes import Int
from fastcs.exceptions import FastCSError
from fastcs.wrappers import command, scan


@pytest.mark.asyncio
async def test_scan_tasks(controller):
    loop = asyncio.get_event_loop()
    transport_options = []
    fastcs = FastCS(controller, transport_options, loop)

    asyncio.create_task(fastcs.serve(interactive=False))
    await asyncio.sleep(0.1)

    for _ in range(3):
        count = controller.count
        await asyncio.sleep(controller.counter.period + 0.01)
        assert controller.count > count


def test_controller_api():
    class MyTestController(Controller):
        attr1: AttrRW[int] = AttrRW(Int())

        def __init__(self):
            super().__init__(description="Controller for testing")

            self.attributes["attr2"] = AttrRW(Int())

        @command()
        async def do_nothing(self):
            pass

        @scan(1.0)
        async def scan_nothing(self):
            pass

    controller = MyTestController()
    api = build_controller_api(controller)

    assert api.description == controller.description
    assert list(api.attributes) == ["attr1", "attr2"]
    assert list(api.command_methods) == ["do_nothing"]
    assert list(api.scan_methods) == ["scan_nothing"]


@pytest.mark.asyncio
async def test_controller_api_methods():
    class MyTestController(Controller):
        def __init__(self):
            super().__init__()

        async def initialise(self):
            async def do_nothing_dynamic() -> None:
                pass

            self.do_nothing_dynamic = Command(do_nothing_dynamic)

        @command()
        async def do_nothing_static(self):
            pass

    controller = MyTestController()
    loop = asyncio.get_event_loop()
    transport_options = []
    fastcs = FastCS(controller, transport_options, loop)

    asyncio.create_task(fastcs.serve(interactive=False))
    await asyncio.sleep(0.1)

    await controller.do_nothing_static()
    await controller.do_nothing_dynamic()

    await fastcs.controller_api.command_methods["do_nothing_static"]()
    await fastcs.controller_api.command_methods["do_nothing_dynamic"]()


@pytest.mark.asyncio
async def test_update_periods():
    @dataclass
    class AttributeIORefTimesCalled(AttributeIORef):
        update_period: float | None = None
        _times_called = 0

    class AttributeIOTimesCalled(AttributeIO[int, AttributeIORefTimesCalled]):
        async def update(self, attr: AttrR[int, AttributeIORefTimesCalled]):
            attr.io_ref._times_called += 1
            await attr.update(attr.io_ref._times_called)

    class MyController(Controller):
        update_once = AttrR(Int(), io_ref=AttributeIORefTimesCalled(update_period=ONCE))
        update_quickly = AttrR(
            Int(), io_ref=AttributeIORefTimesCalled(update_period=0.1)
        )
        update_never = AttrR(
            Int(), io_ref=AttributeIORefTimesCalled(update_period=None)
        )

    controller = MyController(ios=[AttributeIOTimesCalled()])
    loop = asyncio.get_event_loop()
    transport_options = []

    fastcs = FastCS(controller, transport_options, loop)

    assert controller.update_quickly.get() == 0
    assert controller.update_once.get() == 0
    assert controller.update_never.get() == 0

    asyncio.create_task(fastcs.serve(interactive=False))
    await asyncio.sleep(0.5)

    assert controller.update_quickly.get() > 1
    assert controller.update_once.get() == 1
    assert controller.update_never.get() == 0

    assert len(fastcs._scan_tasks) == 1
    assert len(fastcs._initial_coros) == 1


@pytest.mark.asyncio
async def test_scan_raises_exception_via_callback():
    class MyTestController(Controller):
        def __init__(self):
            super().__init__()

        @scan(0.1)
        async def raise_exception(self):
            raise ValueError("Scan Exception")

    controller = MyTestController()
    loop = asyncio.get_event_loop()
    transport_options = []
    fastcs = FastCS(controller, transport_options, loop)

    exception_info = {}
    # This will intercept the exception raised in _scan_done
    loop.set_exception_handler(
        lambda _loop, context: exception_info.update(
            {"exception": context.get("exception")}
        )
    )

    task = asyncio.create_task(fastcs.serve(interactive=False))
    # This allows scan time to run
    await asyncio.sleep(0.2)
    # _scan_done should raise an exception
    assert isinstance(exception_info["exception"], FastCSError)
    for task in fastcs._scan_tasks:
        internal_exception = task.exception()
        assert internal_exception
        # The task exception comes from scan method raise_exception
        assert isinstance(internal_exception, ValueError)
        assert "Scan Exception" == str(internal_exception)


@pytest.mark.asyncio
async def test_controller_connect_disconnect():
    class MyTestController(Controller):
        def __init__(self):
            super().__init__()

            self.connected = False

        async def connect(self):
            self.connected = True

        async def disconnect(self):
            self.connected = False

    controller = MyTestController()

    loop = asyncio.get_event_loop()
    fastcs = FastCS(controller, [], loop)

    task = asyncio.create_task(fastcs.serve(interactive=False))

    # connect is called at the start of serve
    await asyncio.sleep(0.1)
    assert controller.connected

    task.cancel()

    # disconnect is called at the end of serve
    await asyncio.sleep(0.1)
    assert not controller.connected



from enum import IntEnum

import numpy as np
import pytest

from fastcs.datatypes import DataType, Enum, Float, Int, Waveform


def test_base_validate():
    class TestInt(DataType[int]):
        @property
        def dtype(self) -> type[int]:
            return int

    class MyIntEnum(IntEnum):
        A = 0
        B = 1

    test_int = TestInt()

    assert test_int.validate("0") == 0
    assert test_int.validate(MyIntEnum.B) == 1

    with pytest.raises(ValueError, match="Failed to cast"):
        test_int.validate("foo")


@pytest.mark.parametrize(
    ["datatype", "init_args", "value"],
    [
        (Int, {"min": 1}, 0),
        (Int, {"max": -1}, 0),
        (Float, {"min": 1}, 0.0),
        (Float, {"max": -1}, 0.0),
        (Enum, {"enum_cls": int}, 0),
        (Waveform, {"array_dtype": "U64", "shape": (1,)}, np.ndarray([1])),
        (Waveform, {"array_dtype": "float64", "shape": (1, 1)}, np.ndarray([1])),
    ],
)
def test_validate(datatype, init_args, value):
    with pytest.raises(ValueError):
        datatype(**init_args).validate(value)



import asyncio
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, ClassVar, Union

from fastcs.controller_api import ControllerAPI


@dataclass
class Transport:
    """A base class for transport's implementation
    so it can be used in FastCS."""

    subclasses: ClassVar[list[type["Transport"]]] = []

    def __init_subclass__(cls):
        cls.subclasses.append(cls)

    @classmethod
    def union(cls):
        if not cls.subclasses:
            raise RuntimeError("No Transports found")

        return Union[tuple(cls.subclasses)]  # noqa: UP007

    @abstractmethod
    def connect(
        self, controller_api: ControllerAPI, loop: asyncio.AbstractEventLoop
    ) -> None:
        pass

    @property
    def context(self) -> dict[str, Any]:
        return {}

    @abstractmethod
    async def serve(self) -> None:
        pass



import asyncio
from dataclasses import dataclass

from fastcs.tracer import Tracer


class DisconnectedError(Exception):
    """Raised if the ip connection is disconnected."""

    pass


@dataclass
class IPConnectionSettings:
    ip: str = "127.0.0.1"
    port: int = 25565


@dataclass
class StreamConnection:
    """For reading and writing to a stream."""

    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter

    def __post_init__(self):
        self._lock = asyncio.Lock()

    async def __aenter__(self):
        await self._lock.acquire()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()

    async def send_message(self, message) -> None:
        self.writer.write(message.encode("utf-8"))
        await self.writer.drain()

    async def receive_response(self) -> str:
        data = await self.reader.readline()
        return data.decode("utf-8")

    async def close(self):
        self.writer.close()
        await self.writer.wait_closed()


class IPConnection(Tracer):
    """For connecting to an ip using a `StreamConnection`."""

    def __init__(self):
        super().__init__()
        self.__connection = None

    @property
    def _connection(self) -> StreamConnection:
        if self.__connection is None:
            raise DisconnectedError("Need to call connect() before using IPConnection.")

        return self.__connection

    async def connect(self, settings: IPConnectionSettings):
        reader, writer = await asyncio.open_connection(settings.ip, settings.port)
        self.__connection = StreamConnection(reader, writer)

    async def send_command(self, message: str) -> None:
        async with self._connection as connection:
            await connection.send_message(message)

    async def send_query(self, message: str) -> str:
        async with self._connection as connection:
            await connection.send_message(message)
            response = await connection.receive_response()
            self.log_event(
                "Received query response",
                query=message.strip(),
                response=response.strip(),
            )
            return response

    async def close(self):
        async with self._connection as connection:
            await connection.close()
            self.__connection = None



import numpy as np
from pvi.device import (
    LED,
    CheckBox,
    SignalR,
    SignalW,
    TableRead,
    TableWrite,
    TextFormat,
    TextRead,
    TextWrite,
)

from fastcs.attributes import AttrR, AttrW
from fastcs.datatypes import Table
from fastcs.transport.epics.gui import PvaEpicsGUI


def test_get_pv_in_pva(controller_api):
    gui = PvaEpicsGUI(controller_api, "DEVICE")

    assert gui._get_pv([], "A") == "pva://DEVICE:A"
    assert gui._get_pv(["B"], "C") == "pva://DEVICE:B:C"
    assert gui._get_pv(["D", "E"], "F") == "pva://DEVICE:D:E:F"


def test_get_attribute_component_table_write(controller_api):
    gui = PvaEpicsGUI(controller_api, "DEVICE")

    attribute_component = gui._get_attribute_component(
        [],
        "Table",
        AttrW(
            Table(
                structured_dtype=[
                    ("FIELD1", np.uint32),
                    ("FIELD2", np.bool),
                    ("FIELD3", np.dtype("S1000")),
                ]
            )
        ),
    )

    assert isinstance(attribute_component, SignalW)
    assert isinstance(attribute_component.write_widget, TableWrite)
    assert attribute_component.write_widget.widgets == [
        TextWrite(),
        CheckBox(),
        TextWrite(format=TextFormat.string),
    ]


def test_get_attribute_component_table_read(controller_api):
    gui = PvaEpicsGUI(controller_api, "DEVICE")

    attribute_component = gui._get_attribute_component(
        [],
        "Table",
        AttrR(
            Table(
                structured_dtype=[
                    ("FIELD1", np.uint32),
                    ("FIELD2", np.bool),
                    ("FIELD3", np.dtype("S1000")),
                ]
            )
        ),
    )

    assert isinstance(attribute_component, SignalR)
    assert isinstance(attribute_component.read_widget, TableRead)
    assert attribute_component.read_widget.widgets == [
        TextRead(),
        LED(),
        TextRead(format=TextFormat.string),
    ]



import enum

import pytest
from softioc import builder

from fastcs.attributes import AttrRW
from fastcs.datatypes import Bool, Enum, Float, Int, String
from fastcs.transport.epics.ca.util import (
    builder_callable_from_attribute,
    cast_from_epics_type,
    cast_to_epics_type,
    record_metadata_from_datatype,
)


class ShortEnum(enum.Enum):
    NOT = 0
    TOO = 1
    MANY = 2
    VALUES = 3


class LongEnum(enum.Enum):
    THIS = 0
    IS = 1
    AN = 2
    ENUM = 3
    WITH = 4
    ALTOGETHER = 5
    TOO = 6
    MANY = 7
    VALUES = 8
    TO = 9
    BE = 10
    DESCRIBED = 11
    BY = 12
    MBB = 14
    TYPE = 15
    EPICS = 16
    RECORDS = 17


class LongMixedEnum(enum.Enum):
    THIS = "the value is THIS"
    IS = 1
    AN = "the value is AN"
    ENUM = 3
    WITH = "the value is WITH"
    ALTOGETHER = 5
    TOO = "the value is TOO"
    MANY = 7
    VALUES = "the value is VALUES"
    TO = 9
    BE = "the value is BE"
    DESCRIBED = 11
    BY = "the value is BY"
    MBB = 13
    TYPE = "the value is TYPE"
    EPICS = None
    RECORDS = "the value is RECORDS"


class ShortMixedEnum(enum.Enum):
    STRING_MEMBER = "I am a string"
    INT_MEMBER = 2
    NONE_MEMBER = None


@pytest.mark.parametrize(
    "datatype,input,output",
    [
        (Enum(ShortEnum), ShortEnum.TOO, 1),
        # in CA, enums with too many values become epics strings
        (Enum(LongMixedEnum), LongMixedEnum.BE, "BE"),  # string value
        (Enum(LongMixedEnum), LongMixedEnum.EPICS, "EPICS"),  # None value
        (Enum(LongMixedEnum), LongMixedEnum.MBB, "MBB"),  # int value
        (Int(), 4, 4),
        (Float(), 1.0, 1.0),
        (Bool(), True, True),
        (String(), "hey", "hey"),
        # shorter enums can be represented by integers from 0-15
        (Enum(ShortMixedEnum), ShortMixedEnum.STRING_MEMBER, 0),
        (Enum(ShortMixedEnum), ShortMixedEnum.INT_MEMBER, 1),
        (Enum(ShortMixedEnum), ShortMixedEnum.NONE_MEMBER, 2),
    ],
)
def test_casting_to_epics(datatype, input, output):
    assert cast_to_epics_type(datatype, input) == output


@pytest.mark.parametrize(
    "datatype, input",
    [
        # TODO cover Waveform and Table cases
        (Enum(ShortEnum), LongEnum.TOO),  # wrong enum.Enum class
    ],
)
def test_cast_to_epics_validations(datatype, input):
    with pytest.raises(ValueError):
        cast_to_epics_type(datatype, input)


@pytest.mark.parametrize(
    "datatype,from_epics,result",
    [
        # long enums backed by strings
        (Enum(LongMixedEnum), "BE", LongMixedEnum.BE),  # string value
        (Enum(LongMixedEnum), "EPICS", LongMixedEnum.EPICS),  # None value
        (Enum(LongMixedEnum), "MBB", LongMixedEnum.MBB),  # int value
        (Int(), 4, 4),
        (Float(), 1.0, 1.0),
        (Bool(), True, True),
        (String(), "hey", "hey"),
        (Enum(ShortEnum), 2, ShortEnum.MANY),
        # short enums backed by mbbi/mbbo
        (Enum(ShortMixedEnum), 0, ShortMixedEnum.STRING_MEMBER),
        (Enum(ShortMixedEnum), 1, ShortMixedEnum.INT_MEMBER),
        (Enum(ShortMixedEnum), 2, ShortMixedEnum.NONE_MEMBER),
        (Bool(), 1, True),
        (Bool(), 0, False),
    ],
)
def test_cast_from_epics_type(datatype, from_epics, result):
    assert cast_from_epics_type(datatype, from_epics) == result


@pytest.mark.parametrize(
    "datatype, input",
    [
        (object(), 0),
        (Bool(), 3),
    ],
)
def test_cast_from_epics_validations(datatype, input):
    with pytest.raises(ValueError):
        cast_from_epics_type(datatype, input)


@pytest.mark.parametrize(
    "datatype,in_record,out_record",
    [
        (Enum(ShortEnum), builder.mbbIn, builder.mbbOut),
        # long enums use string even if all values are ints
        (Enum(LongEnum), builder.longStringIn, builder.longStringOut),
        (Enum(LongMixedEnum), builder.longStringIn, builder.longStringOut),
    ],
)
def test_builder_callable_enum_types(datatype, in_record, out_record):
    attr = AttrRW(datatype)
    assert builder_callable_from_attribute(attr, False) == out_record
    assert builder_callable_from_attribute(attr, True) == in_record


def test_drive_metadata_from_datatype():
    dtype = Float(units="mm", min=-10.0, max=10.0, min_alarm=-5, max_alarm=5, prec=3)
    out_arguments = record_metadata_from_datatype(dtype, True)
    assert out_arguments == {
        "DRVH": 10.0,
        "DRVL": -10.0,
        "EGU": "mm",
        "HOPR": 5,
        "LOPR": -5,
        "PREC": 3,
    }
    in_arguments = record_metadata_from_datatype(dtype, False)
    assert in_arguments == {"EGU": "mm", "HOPR": 5, "LOPR": -5, "PREC": 3}



import numpy as np
from p4p import Value
from p4p.nt import NTEnum, NTNDArray, NTScalar, NTTable
from p4p.nt.enum import ntenum
from p4p.nt.ndarray import ntndarray
from p4p.server import ServerOperation
from p4p.server.asyncio import SharedPV

from fastcs.attributes import Attribute, AttrR, AttrRW, AttrW
from fastcs.cs_methods import CommandCallback
from fastcs.datatypes import Table
from fastcs.tracer import Tracer

from .types import (
    MAJOR_ALARM_SEVERITY,
    RECORD_ALARM_STATUS,
    cast_from_p4p_value,
    cast_to_p4p_value,
    make_p4p_type,
    p4p_alarm_states,
    p4p_timestamp_now,
)

tracer = Tracer(name=__name__)


class WritePvHandler:
    def __init__(self, attr_w: AttrW | AttrRW):
        self._attr_w = attr_w

    async def put(self, pv: SharedPV, op: ServerOperation):
        value = op.value()
        if isinstance(self._attr_w.datatype, Table):
            assert isinstance(value, list)
            raw_value = np.array(
                [tuple(labelled_row.values()) for labelled_row in value],
                dtype=self._attr_w.datatype.structured_dtype,
            )
        elif isinstance(value, Value):
            raw_value = value.todict()["value"]
        else:
            # Unfortunately these types don't have a `todict`,
            # while our `buildType` fields don't have a `.raw`.
            assert isinstance(value, ntenum | ntndarray)
            raw_value = value.raw.value  # type:ignore

        cast_value = cast_from_p4p_value(self._attr_w, raw_value)

        tracer.log_event("PV put", topic=self._attr_w, pv=pv, value=cast_value)
        await self._attr_w.put(cast_value)
        op.done()


class CommandPvHandler:
    def __init__(self, command: CommandCallback):
        self._command = command
        self._task_in_progress = False

    async def _run_command(self) -> dict:
        self._task_in_progress = True

        try:
            await self._command()
        except Exception as e:
            alarm_states = p4p_alarm_states(
                MAJOR_ALARM_SEVERITY, RECORD_ALARM_STATUS, str(e)
            )
        else:
            alarm_states = p4p_alarm_states()

        self._task_in_progress = False
        return alarm_states

    async def put(self, pv: SharedPV, op: ServerOperation):
        value = op.value()
        raw_value = value["value"]

        if raw_value is True:
            if self._task_in_progress:
                raise RuntimeError(
                    "Received request to run command but it is already in progress. "
                    "Maybe the command should spawn an asyncio task?"
                )

            # Check if record block request recieved
            match op.pvRequest().todict():
                case {"record": {"_options": {"block": "true"}}}:
                    blocking = True
                case _:
                    blocking = False

            # Flip to true once command task starts
            pv.post({"value": True, **p4p_timestamp_now(), **p4p_alarm_states()})
            if not blocking:
                op.done()
            alarm_states = await self._run_command()
            pv.post({"value": False, **p4p_timestamp_now(), **alarm_states})
            if blocking:
                op.done()
        else:
            raise RuntimeError("Commands should only take the value `True`.")


def _make_shared_pv_arguments(attribute: Attribute) -> dict[str, object]:
    type_ = make_p4p_type(attribute)
    if isinstance(type_, (NTEnum | NTNDArray | NTTable)):
        return {"nt": type_}
    else:

        def _wrap(value: dict):
            return Value(type_, value)

        return {"wrap": _wrap}


def make_shared_read_pv(attribute: AttrR) -> SharedPV:
    shared_pv = SharedPV(
        initial=cast_to_p4p_value(attribute, attribute.get()),
        **_make_shared_pv_arguments(attribute),
    )

    async def set_readback(value):
        tracer.log_event("PV set readback", topic=attribute, value=value)
        shared_pv.post(cast_to_p4p_value(attribute, value))

    attribute.add_on_update_callback(set_readback)

    return shared_pv


def make_shared_write_pv(attribute: AttrW) -> SharedPV:
    shared_pv = SharedPV(
        handler=WritePvHandler(attribute),
        initial=cast_to_p4p_value(attribute, attribute.datatype.initial_value),
        **_make_shared_pv_arguments(attribute),
    )

    async def set_setpoint(value):
        tracer.log_event("PV set setpoint", topic=attribute, value=value)
        shared_pv.post(cast_to_p4p_value(attribute, value))

    attribute.add_sync_setpoint_callback(set_setpoint)

    return shared_pv


def make_command_pv(command: CommandCallback) -> SharedPV:
    type_ = NTScalar.buildType("?", display=True, control=True)

    initial = Value(type_, {"value": False, **p4p_alarm_states()})

    def _wrap(value: dict):
        return Value(type_, value)

    shared_pv = SharedPV(
        initial=initial,
        handler=CommandPvHandler(command),
        wrap=_wrap,
    )

    return shared_pv



import math
import time

import numpy as np
from numpy.typing import DTypeLike
from p4p import Value
from p4p.nt import NTEnum, NTNDArray, NTScalar, NTTable

from fastcs.attributes import Attribute, AttrR, AttrW
from fastcs.datatypes import Bool, Enum, Float, Int, String, T, Table, Waveform

P4P_ALLOWED_DATATYPES = (Int, Float, String, Bool, Enum, Waveform, Table)

# https://epics-base.github.io/pvxs/nt.html#alarm-t
RECORD_ALARM_STATUS = 3
NO_ALARM_STATUS = 0
MAJOR_ALARM_SEVERITY = 2
NO_ALARM_SEVERITY = 0

# https://numpy.org/devdocs/reference/arrays.dtypes.html#arrays-dtypes
# Some numpy dtypes don't match directly with the p4p ones
_NUMPY_DTYPE_TO_P4P_DTYPE = {
    "S": "s",  # Raw bytes to unicode bytes
    "U": "s",
}


def _table_with_numpy_dtypes_to_p4p_dtypes(numpy_dtypes: list[tuple[str, DTypeLike]]):
    """
    Numpy structured datatypes can use the numpy dtype class, e.g `np.int32` or the
    character, e.g "i". P4P only accepts the character so this method is used to
    convert.

    https://epics-base.github.io/p4p/values.html#type-definitions

    It also forbids:
        The numpy dtype for float16, which isn't supported in p4p.
        String types which should be supported but currently don't function:
            https://github.com/epics-base/p4p/issues/168
    """
    p4p_dtypes = []
    for name, numpy_dtype in numpy_dtypes:
        dtype_char = np.dtype(numpy_dtype).char
        dtype_char = _NUMPY_DTYPE_TO_P4P_DTYPE.get(dtype_char, dtype_char)
        if dtype_char in ("e", "U", "S"):
            raise ValueError(f"`{np.dtype(numpy_dtype)}` is unsupported in p4p.")
        p4p_dtypes.append((name, dtype_char))
    return p4p_dtypes


def make_p4p_type(
    attribute: Attribute,
) -> NTScalar | NTEnum | NTNDArray | NTTable:
    """Creates a p4p type for a given `Attribute` s `fastcs.datatypes.DataType`."""
    display = isinstance(attribute, AttrR)
    control = isinstance(attribute, AttrW)
    match attribute.datatype:
        case Int():
            return NTScalar.buildType("i", display=display, control=control)
        case Float():
            return NTScalar.buildType("d", display=display, control=control, form=True)
        case String():
            return NTScalar.buildType("s", display=display, control=control)
        case Bool():
            return NTScalar.buildType("?", display=display, control=control)
        case Enum():
            return NTEnum()
        case Waveform():
            # TODO: https://github.com/DiamondLightSource/FastCS/issues/123
            # * Make 1D scalar array for 1D shapes.
            #     This will require converting from np.int32 to "ai"
            #     if len(shape) == 1:
            #         return NTScalarArray(convert np.datatype32 to string "ad")
            # * Add an option for allowing shape to change, if so we will
            #   use an NDArray here even if shape is 1D

            return NTNDArray()
        case Table(structured_dtype):
            # TODO: `NTEnum/NTNDArray/NTTable.wrap` don't accept extra fields until
            # https://github.com/epics-base/p4p/issues/166
            return NTTable(
                columns=_table_with_numpy_dtypes_to_p4p_dtypes(structured_dtype)
            )
        case _:
            raise RuntimeError(f"DataType `{attribute.datatype}` unsupported in P4P.")


def cast_from_p4p_value(attribute: Attribute[T], value: object) -> T:
    """Converts from a p4p value to a FastCS `Attribute` value."""
    match attribute.datatype:
        case Enum():
            return attribute.datatype.validate(attribute.datatype.members[value.index])
        case Waveform(shape=shape):
            # p4p sends a flattened array
            assert value.shape == (math.prod(shape),)
            return attribute.datatype.validate(value.reshape(attribute.datatype.shape))
        case Table(structured_dtype):
            assert isinstance(value, np.ndarray)
            return attribute.datatype.validate(np.array(value, dtype=structured_dtype))
        case attribute.datatype if issubclass(
            type(attribute.datatype), P4P_ALLOWED_DATATYPES
        ):
            return attribute.datatype.validate(value)  # type: ignore
        case _:
            raise ValueError(f"Unsupported datatype {attribute.datatype}")


def p4p_alarm_states(
    severity: int = NO_ALARM_SEVERITY,
    status: int = NO_ALARM_STATUS,
    message: str = "",
) -> dict:
    """Returns the p4p alarm structure for a given severity, status, and message."""
    return {
        "alarm": {
            "severity": severity,
            "status": status,
            "message": message,
        },
    }


def p4p_timestamp_now() -> dict:
    """The p4p timestamp structure for the current time."""
    now = time.time()
    seconds_past_epoch = int(now)
    nanoseconds = int((now - seconds_past_epoch) * 1e9)
    return {
        "timeStamp": {
            "secondsPastEpoch": seconds_past_epoch,
            "nanoseconds": nanoseconds,
        }
    }


def p4p_display(attribute: Attribute) -> dict:
    """Gets the p4p display structure for a given attribute."""
    display = {}
    if attribute.description is not None:
        display["description"] = attribute.description
    if isinstance(attribute.datatype, (Float | Int)):
        if attribute.datatype.max is not None:
            display["limitHigh"] = attribute.datatype.max
        if attribute.datatype.min is not None:
            display["limitLow"] = attribute.datatype.min
        if attribute.datatype.units is not None:
            display["units"] = attribute.datatype.units
    if isinstance(attribute.datatype, Float):
        if attribute.datatype.prec is not None:
            display["precision"] = attribute.datatype.prec
    if display:
        return {"display": display}
    return {}


def _p4p_check_numerical_for_alarm_states(datatype: Int | Float, value: T) -> dict:
    low = None if datatype.min_alarm is None else value < datatype.min_alarm  # type: ignore
    high = None if datatype.max_alarm is None else value > datatype.max_alarm  # type: ignore
    severity = (
        MAJOR_ALARM_SEVERITY
        if high not in (None, False) or low not in (None, False)
        else NO_ALARM_SEVERITY
    )
    status, message = NO_ALARM_SEVERITY, "No alarm"
    if low:
        status, message = (
            RECORD_ALARM_STATUS,
            f"Below minimum alarm limit: {datatype.min_alarm}",
        )
    if high:
        status, message = (
            RECORD_ALARM_STATUS,
            f"Above maximum alarm limit: {datatype.max_alarm}",
        )

    return p4p_alarm_states(severity, status, message)


def cast_to_p4p_value(attribute: Attribute[T], value: T) -> object:
    """Converts a FastCS `Attribute` value to a p4p value,
    including metadata and alarm states."""
    match attribute.datatype:
        case Enum():
            return {
                "index": attribute.datatype.index_of(value),
                "choices": attribute.datatype.names,
            }
        case Waveform():
            return attribute.datatype.validate(value)
        case Table():
            return attribute.datatype.validate(value)

        case datatype if issubclass(type(datatype), P4P_ALLOWED_DATATYPES):
            record_fields: dict = {"value": datatype.validate(value)}
            if isinstance(attribute, AttrR):
                record_fields.update(p4p_display(attribute))

            if isinstance(datatype, (Float | Int)):
                record_fields.update(
                    _p4p_check_numerical_for_alarm_states(
                        datatype,
                        value,
                    )
                )
            else:
                record_fields.update(p4p_alarm_states())

            record_fields.update(p4p_timestamp_now())

            return Value(make_p4p_type(attribute), record_fields)
        case _:
            raise ValueError(f"Unsupported datatype {attribute.datatype}")



import json
import os
from dataclasses import dataclass

import pytest
from pydantic import create_model
from pytest_mock import MockerFixture
from ruamel.yaml import YAML
from typer.testing import CliRunner

from fastcs import __version__
from fastcs.attributes import AttrR
from fastcs.controller import Controller
from fastcs.datatypes import Int
from fastcs.exceptions import LaunchError
from fastcs.launch import (
    _launch,
    get_controller_schema,
    launch,
)
from fastcs.transport.transport import Transport


@dataclass
class SomeConfig:
    name: str


class SingleArg(Controller):
    def __init__(self):
        super().__init__()


class NotHinted(Controller):
    def __init__(self, arg):
        super().__init__()


class IsHinted(Controller):
    read = AttrR(Int())

    def __init__(self, arg: SomeConfig) -> None:
        super().__init__()


class ManyArgs(Controller):
    def __init__(self, arg: SomeConfig, too_many):
        super().__init__()


runner = CliRunner()


def test_single_arg_schema():
    target_model = create_model(
        "SingleArg",
        transport=(list[Transport.union()], ...),
        __config__={"extra": "forbid"},
    )
    target_dict = target_model.model_json_schema()

    app = _launch(SingleArg)
    result = runner.invoke(app, ["schema"])
    assert result.exit_code == 0
    result_dict = json.loads(result.stdout)

    assert result_dict == target_dict


def test_is_hinted_schema(data):
    target_model = create_model(
        "IsHinted",
        controller=(SomeConfig, ...),
        transport=(list[Transport.union()], ...),
        __config__={"extra": "forbid"},
    )
    target_dict = target_model.model_json_schema()

    app = _launch(IsHinted)
    result = runner.invoke(app, ["schema"])
    assert result.exit_code == 0
    result_dict = json.loads(result.stdout)

    assert result_dict == target_dict

    # # store a schema to use for debugging
    # with open(data / "schema.json", mode="w") as f:
    #     json.dump(result_dict, f, indent=2)


def test_not_hinted_schema():
    error = (
        "Expected typehinting in 'NotHinted.__init__' but received "
        "(self, arg). Add a typehint for `arg`."
    )

    with pytest.raises(LaunchError) as exc_info:
        launch(NotHinted)
    assert str(exc_info.value) == error


def test_over_defined_schema():
    error = (
        ""
        "Expected no more than 2 arguments for 'ManyArgs.__init__' "
        "but received 3 as `(self, arg: tests.test_launch.SomeConfig, too_many)`"
    )

    with pytest.raises(LaunchError) as exc_info:
        launch(ManyArgs)
    assert str(exc_info.value) == error


def test_version():
    impl_version = "0.0.1"
    expected = f"SingleArg: {impl_version}\nFastCS: {__version__}\n"
    app = _launch(SingleArg, version=impl_version)
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert result.stdout == expected


def test_no_version():
    expected = f"FastCS: {__version__}\n"
    app = _launch(SingleArg)
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert result.stdout == expected


def test_launch(mocker: MockerFixture, data):
    run = mocker.patch("fastcs.launch.FastCS.run")

    app = _launch(IsHinted)
    result = runner.invoke(app, ["run", str(data / "config.yaml")])
    assert result.exit_code == 0

    run.assert_called_once()


def test_get_schema(data):
    target_schema = get_controller_schema(IsHinted)
    if os.environ.get("FASTCS_REGENERATE_OUTPUT", None):
        with open(data / "schema.json", "w") as f:
            json.dump(target_schema, f, indent=2)

    ref_schema = YAML(typ="safe").load(data / "schema.json")
    assert target_schema == ref_schema


def test_error_if_identical_context_in_transports(mocker: MockerFixture, data):
    mocker.patch(
        "fastcs.transport.Transport.context",
        new_callable=mocker.PropertyMock,
        return_value={"controller": "test"},
    )
    app = _launch(IsHinted)
    result = runner.invoke(app, ["run", str(data / "config.yaml")])
    assert isinstance(result.exception, RuntimeError)
    assert "Duplicate context keys found" in result.exception.args[0]



import pytest

from fastcs.controller import Controller
from fastcs.cs_methods import Command, Method, Scan, UnboundCommand, UnboundScan
from fastcs.exceptions import FastCSError


def test_method():
    def sync_do_nothing():
        pass

    with pytest.raises(FastCSError):
        Method(sync_do_nothing)  # type: ignore

    async def do_nothing_with_return() -> int:
        return 1

    with pytest.raises(FastCSError):
        Method(do_nothing_with_return)  # type: ignore

    async def do_nothing():
        """Do nothing."""
        pass

    method = Method(do_nothing, group="Nothing")

    assert method.docstring == "Do nothing."
    assert method.group == "Nothing"


@pytest.mark.asyncio
async def test_unbound_command():
    class TestController(Controller):
        async def do_nothing(self):
            pass

        async def do_nothing_with_arg(self, arg):
            pass

    unbound_command = UnboundCommand(TestController.do_nothing, group="Test")

    with pytest.raises(NotImplementedError):
        await unbound_command()

    with pytest.raises(FastCSError):
        UnboundCommand(TestController.do_nothing_with_arg)  # type: ignore

    with pytest.raises(FastCSError):
        Command(TestController().do_nothing_with_arg)  # type: ignore

    command = unbound_command.bind(TestController())
    # Test that group is passed when binding commands
    assert command.group == "Test"

    await command()


@pytest.mark.asyncio
async def test_unbound_scan():
    class TestController(Controller):
        async def update_nothing(self):
            pass

        async def update_nothing_with_arg(self, arg):
            pass

    unbound_scan = UnboundScan(TestController.update_nothing, 1.0)

    assert unbound_scan.period == 1.0

    with pytest.raises(NotImplementedError):
        await unbound_scan()

    with pytest.raises(FastCSError):
        UnboundScan(TestController.update_nothing_with_arg, 1.0)  # type: ignore

    with pytest.raises(FastCSError):
        Scan(TestController().update_nothing_with_arg, 1.0)  # type: ignore

    scan = unbound_scan.bind(TestController())

    assert scan.period == 1.0

    await scan()



import asyncio
import enum

import numpy as np

from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.controller import Controller
from fastcs.datatypes import Bool, Enum, Float, Int, Table, Waveform
from fastcs.launch import FastCS
from fastcs.transport.epics.options import (
    EpicsIOCOptions,
)
from fastcs.transport.epics.pva.transport import EpicsPVATransport
from fastcs.wrappers import command, scan


class FEnum(enum.Enum):
    A = 0
    B = 1
    C = "VALUES ARE ARBITRARY"
    D = 2
    E = 5


class ParentController(Controller):
    description = "some controller"
    a: AttrRW = AttrRW(Int(max=400_000, max_alarm=40_000))
    b: AttrW = AttrW(Float(min=-1, min_alarm=-0.5))

    table: AttrRW = AttrRW(
        Table([("A", np.int32), ("B", "i"), ("C", "?"), ("D", np.float64)])
    )


class ChildController(Controller):
    fail_on_next_e = True
    c: AttrW = AttrW(Int())

    @command()
    async def d(self):
        print("D: RUNNING")
        await asyncio.sleep(0.1)
        print("D: FINISHED")
        await self.j.update(self.j.get() + 1)

    e: AttrR = AttrR(Bool())

    @scan(1)
    async def flip_flop(self):
        await self.e.update(not self.e.get())

    f: AttrRW = AttrRW(Enum(FEnum))
    g: AttrRW = AttrRW(Waveform(np.int64, shape=(3,)))
    h: AttrRW = AttrRW(Waveform(np.float64, shape=(3, 3)))

    @command()
    async def i(self):
        print("I: RUNNING")
        await asyncio.sleep(0.1)
        if self.fail_on_next_e:
            self.fail_on_next_e = False
            raise RuntimeError("I: FAILED WITH THIS WEIRD ERROR")
        else:
            self.fail_on_next_e = True
            print("I: FINISHED")
            await self.j.update(self.j.get() + 1)

    j: AttrR = AttrR(Int())


def run(pv_prefix="P4P_TEST_DEVICE"):
    controller = ParentController()
    controller.a.enable_tracing()
    controller.child1 = ChildController(description="some sub controller")
    controller.child2 = ChildController(description="another sub controller")

    fastcs = FastCS(
        controller, [EpicsPVATransport(pva_ioc=EpicsIOCOptions(pv_prefix=pv_prefix))]
    )
    fastcs.run(interactive=False)


if __name__ == "__main__":
    run()



from multiprocessing import Queue

from p4p import Value
from p4p.client.thread import Context


def test_ioc(softioc_subprocess: tuple[str, Queue]):
    pv_prefix, _ = softioc_subprocess
    ctxt = Context("pva")

    _parent_pvi = ctxt.get(f"{pv_prefix}:PVI")
    assert isinstance(_parent_pvi, Value)
    parent_pvi = _parent_pvi.todict()
    assert all(f in parent_pvi for f in ("alarm", "display", "timeStamp", "value"))
    assert parent_pvi["display"] == {"description": "The records in this controller"}
    assert parent_pvi["value"] == {
        "a": {"r": f"{pv_prefix}:A"},
        "b": {"r": f"{pv_prefix}:B_RBV", "w": f"{pv_prefix}:B"},
        "child": {"d": f"{pv_prefix}:Child:PVI"},
    }

    child_pvi_pv = parent_pvi["value"]["child"]["d"]
    _child_pvi = ctxt.get(child_pvi_pv)
    assert isinstance(_child_pvi, Value)
    child_pvi = _child_pvi.todict()
    assert all(f in child_pvi for f in ("alarm", "display", "timeStamp", "value"))
    assert child_pvi["display"] == {"description": "The records in this controller"}
    assert child_pvi["value"] == {
        "c": {"w": f"{pv_prefix}:Child:C"},
        "d": {"x": f"{pv_prefix}:Child:D"},
    }



import glob
import runpy
import subprocess
import time
from pathlib import Path

import pytest

HERE = Path(__file__).parent


@pytest.fixture(scope="module", autouse=True)
def sim_temperature_controller():
    config_path: str = f"{HERE}/../src/fastcs/demo/simulation/temp_controller.yaml"
    process = subprocess.Popen(
        ["tickit", "all", config_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )

    timeout = 10
    start_time = time.monotonic()
    while process.stdout is not None:
        line = process.stdout.readline()
        if "Temperature controller running" in line:
            break

        if time.monotonic() - start_time > timeout:
            raise TimeoutError("Simulator did not start in time")

        time.sleep(0.1)

    yield

    process.kill()
    print(process.communicate()[0])


@pytest.mark.parametrize("filename", glob.glob("docs/snippets/*.py", recursive=True))
def test_snippet(filename):
    runpy.run_path(filename)



from collections.abc import Awaitable, Callable, Coroutine
from typing import Any

import strawberry
import uvicorn
from strawberry.asgi import GraphQL
from strawberry.tools import create_type
from strawberry.types.field import StrawberryField

from fastcs.attributes import AttrR, AttrRW, AttrW, T
from fastcs.controller_api import ControllerAPI
from fastcs.exceptions import FastCSError
from fastcs.logging import intercept_std_logger

from .options import GraphQLServerOptions


class GraphQLServer:
    """A GraphQL server which handles a controller"""

    def __init__(self, controller_api: ControllerAPI):
        self._controller_api = controller_api
        self._app = self._create_app()

    def _create_app(self) -> GraphQL:
        api = GraphQLAPI(self._controller_api)
        schema = api.create_schema()
        app = GraphQL(schema)

        return app

    async def serve(self, options: GraphQLServerOptions | None = None) -> None:
        options = options or GraphQLServerOptions()
        self._server = uvicorn.Server(
            uvicorn.Config(
                app=self._app,
                host=options.host,
                port=options.port,
                log_level=options.log_level,
            )
        )
        intercept_std_logger("uvicorn.error")
        intercept_std_logger("uvicorn.access")
        intercept_std_logger("uvicorn.asgi")

        await self._server.serve()


class GraphQLAPI:
    """A Strawberry API built dynamically from a `ControllerAPI`"""

    def __init__(self, controller_api: ControllerAPI):
        self.queries: list[StrawberryField] = []
        self.mutations: list[StrawberryField] = []

        self._process_attributes(controller_api)
        self._process_commands(controller_api)
        self._process_sub_apis(controller_api)

    def _process_attributes(self, api: ControllerAPI):
        """Create queries and mutations from api attributes."""
        for attr_name, attribute in api.attributes.items():
            match attribute:
                # mutation for server changes https://graphql.org/learn/queries/
                case AttrRW():
                    self.queries.append(
                        strawberry.field(_wrap_attr_get(attr_name, attribute))
                    )
                    self.mutations.append(
                        strawberry.mutation(_wrap_attr_set(attr_name, attribute))
                    )
                case AttrR():
                    self.queries.append(
                        strawberry.field(_wrap_attr_get(attr_name, attribute))
                    )
                case AttrW():
                    self.mutations.append(
                        strawberry.mutation(_wrap_attr_set(attr_name, attribute))
                    )

    def _process_commands(self, controller_api: ControllerAPI):
        """Create mutations from api commands"""
        for name, method in controller_api.command_methods.items():
            self.mutations.append(strawberry.mutation(_wrap_command(name, method.fn)))

    def _process_sub_apis(self, root_controller_api: ControllerAPI):
        """Recursively add fields from the queries and mutations of sub apis"""
        for controller_api in root_controller_api.sub_apis.values():
            name = "".join(controller_api.path)
            child_tree = GraphQLAPI(controller_api)
            if child_tree.queries:
                self.queries.append(
                    _wrap_as_field(
                        name, create_type(f"{name}Query", child_tree.queries)
                    )
                )
            if child_tree.mutations:
                self.mutations.append(
                    _wrap_as_field(
                        name, create_type(f"{name}Mutation", child_tree.mutations)
                    )
                )

    def create_schema(self) -> strawberry.Schema:
        """Create a Strawberry Schema to load into a GraphQL application."""
        if not self.queries:
            raise FastCSError(
                "Can't create GraphQL transport from ControllerAPI with no read "
                "attributes"
            )

        query = create_type("Query", self.queries)
        mutation = create_type("Mutation", self.mutations) if self.mutations else None

        return strawberry.Schema(query=query, mutation=mutation)


def _wrap_attr_set(
    attr_name: str, attribute: AttrW[T]
) -> Callable[[T], Coroutine[Any, Any, None]]:
    """Wrap an attribute in a function with annotations for strawberry"""

    async def _dynamic_f(value):
        await attribute.put(value)
        return value

    # Add type annotations for validation, schema, conversions
    _dynamic_f.__name__ = attr_name
    _dynamic_f.__annotations__["value"] = attribute.datatype.dtype
    _dynamic_f.__annotations__["return"] = attribute.datatype.dtype

    return _dynamic_f


def _wrap_attr_get(
    attr_name: str, attribute: AttrR[T]
) -> Callable[[], Coroutine[Any, Any, Any]]:
    """Wrap an attribute in a function with annotations for strawberry"""

    async def _dynamic_f() -> Any:
        return attribute.get()

    _dynamic_f.__name__ = attr_name
    _dynamic_f.__annotations__["return"] = attribute.datatype.dtype

    return _dynamic_f


def _wrap_as_field(field_name: str, operation: type) -> StrawberryField:
    """Wrap a strawberry type as a field of a parent type"""

    def _dynamic_field():
        return operation()

    _dynamic_field.__name__ = field_name
    _dynamic_field.__annotations__["return"] = operation

    return strawberry.field(_dynamic_field)


def _wrap_command(method_name: str, method: Callable) -> Callable[..., Awaitable[bool]]:
    """Wrap a command in a function with annotations for strawberry"""

    async def _dynamic_f() -> bool:
        await method()
        return True

    _dynamic_f.__name__ = method_name

    return _dynamic_f



import copy
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Literal

from pytest_mock import MockerFixture, MockType

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.control_system import build_controller_api
from fastcs.controller import Controller
from fastcs.controller_api import ControllerAPI
from fastcs.datatypes import Int, T
from fastcs.wrappers import command, scan


@dataclass
class MyTestAttributeIORef(AttributeIORef):
    update_period = 1


class MyTestAttributeIO(AttributeIO[T, MyTestAttributeIORef]):
    async def update(self, attr: AttrR[T, MyTestAttributeIORef]):
        print(f"update {attr}")

    async def send(self, attr: AttrW[T, MyTestAttributeIORef], value: T):
        print(f"sending {attr} = {value}")
        if isinstance(attr, AttrRW):
            await attr.update(value)


test_attribute_io = MyTestAttributeIO()  # instance


class TestSubController(Controller):
    read_int: AttrR = AttrR(Int(), io_ref=MyTestAttributeIORef())

    def __init__(self) -> None:
        super().__init__(ios=[test_attribute_io])


class MyTestController(Controller):
    def __init__(self) -> None:
        super().__init__(ios=[test_attribute_io])

        self._sub_controllers: list[TestSubController] = []
        for index in range(1, 3):
            controller = TestSubController()
            self._sub_controllers.append(controller)
            self.add_sub_controller(f"SubController{index:02d}", controller)

    initialised = False
    connected = False
    count = 0

    async def initialise(self) -> None:
        await super().initialise()
        self.initialised = True

    async def connect(self) -> None:
        self.connected = True

    async def disconnect(self) -> None:
        self.connected = False

    @command()
    async def go(self):
        pass

    @scan(0.01)
    async def counter(self):
        self.count += 1


class AssertableControllerAPI(ControllerAPI):
    def __init__(self, controller: Controller, mocker: MockerFixture) -> None:
        super().__init__()

        self.mocker = mocker
        self.command_method_spys: dict[str, MockType] = {}

        # Build a ControllerAPI from the given Controller
        controller_api = build_controller_api(controller)
        # Copy its fields
        self.attributes = controller_api.attributes
        self.command_methods = controller_api.command_methods
        self.scan_methods = controller_api.scan_methods
        self.sub_apis = controller_api.sub_apis

        # Create spys for command methods before they are passed to the transport
        for command_name in self.command_methods.keys():
            self.command_method_spys[command_name] = mocker.spy(
                self.command_methods[command_name], "_fn"
            )

    @contextmanager
    def assert_read_here(self, path: list[str]):
        yield from self._assert_method(path, "get")

    @contextmanager
    def assert_write_here(self, path: list[str]):
        yield from self._assert_method(path, "put")

    @contextmanager
    def assert_execute_here(self, path: list[str]):
        yield from self._assert_method(path, "")

    def _assert_method(self, path: list[str], method: Literal["get", "put", ""]):
        """
        This context manager can be used to confirm that a fastcs
        controller's respective attribute or command methods are called
        a single time within a context block
        """
        queue = copy.deepcopy(path)

        # Navigate to sub controller
        controller_api = self
        item_name = queue.pop(-1)
        for item in queue:
            controller_api = controller_api.sub_apis[item]

        # Get spy
        if method:
            attr = controller_api.attributes[item_name]
            spy = self.mocker.spy(attr, method)
        else:
            # Lookup pre-defined spy for method
            spy = self.command_method_spys[item_name]

        initial = spy.call_count
        try:
            yield  # Enter context
        except Exception as e:
            raise e
        else:  # Exit context
            final = spy.call_count
            assert final == initial + 1, (
                f"Expected {'.'.join(path + [method] if method else path)} "
                f"to be called once, but it was called {final - initial} times."
            )



import enum
from dataclasses import KW_ONLY, dataclass
from pathlib import Path
from typing import TypeVar

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.connections import IPConnection, IPConnectionSettings
from fastcs.controller import Controller
from fastcs.datatypes import Enum, Float, Int, String
from fastcs.launch import FastCS
from fastcs.transport.epics.ca.transport import EpicsCATransport
from fastcs.transport.epics.options import EpicsGUIOptions, EpicsIOCOptions

NumberT = TypeVar("NumberT", int, float)


@dataclass
class TemperatureControllerAttributeIORef(AttributeIORef):
    name: str
    _: KW_ONLY
    update_period: float | None = 0.2


class TemperatureControllerAttributeIO(
    AttributeIO[NumberT, TemperatureControllerAttributeIORef]
):
    def __init__(self, connection: IPConnection, suffix: str = ""):
        super().__init__()

        self._connection = connection
        self._suffix = suffix

    async def update(self, attr: AttrR[NumberT, TemperatureControllerAttributeIORef]):
        query = f"{attr.io_ref.name}{self._suffix}?"
        response = await self._connection.send_query(f"{query}\r\n")
        value = response.strip("\r\n")

        await attr.update(attr.dtype(value))

    async def send(
        self, attr: AttrW[NumberT, TemperatureControllerAttributeIORef], value: NumberT
    ) -> None:
        command = f"{attr.io_ref.name}{self._suffix}={attr.dtype(value)}"
        await self._connection.send_command(f"{command}\r\n")


class OnOffEnum(enum.StrEnum):
    Off = "0"
    On = "1"


class TemperatureRampController(Controller):
    start = AttrRW(Int(), io_ref=TemperatureControllerAttributeIORef(name="S"))
    end = AttrRW(Int(), io_ref=TemperatureControllerAttributeIORef(name="E"))
    enabled = AttrRW(Enum(OnOffEnum), io_ref=TemperatureControllerAttributeIORef("N"))

    def __init__(self, index: int, connection: IPConnection) -> None:
        suffix = f"{index:02d}"
        super().__init__(
            f"Ramp{suffix}", ios=[TemperatureControllerAttributeIO(connection, suffix)]
        )


class TemperatureController(Controller):
    device_id = AttrR(String(), io_ref=TemperatureControllerAttributeIORef("ID"))
    power = AttrR(Float(), io_ref=TemperatureControllerAttributeIORef("P"))
    ramp_rate = AttrRW(Float(), io_ref=TemperatureControllerAttributeIORef("R"))

    def __init__(self, ramp_count: int, settings: IPConnectionSettings):
        self._ip_settings = settings
        self._connection = IPConnection()

        super().__init__(ios=[TemperatureControllerAttributeIO(self._connection)])

        self._ramp_controllers: list[TemperatureRampController] = []
        for index in range(1, ramp_count + 1):
            controller = TemperatureRampController(index, self._connection)
            self._ramp_controllers.append(controller)
            self.add_sub_controller(f"R{index}", controller)

    async def connect(self):
        await self._connection.connect(self._ip_settings)


gui_options = EpicsGUIOptions(
    output_path=Path(".") / "demo.bob", title="Demo Temperature Controller"
)
epics_ca = EpicsCATransport(gui=gui_options, ca_ioc=EpicsIOCOptions(pv_prefix="DEMO"))
connection_settings = IPConnectionSettings("localhost", 25565)
fastcs = FastCS(TemperatureController(4, connection_settings), [epics_ca])

if __name__ == "__main__":
    fastcs.run()



from __future__ import annotations

from collections import Counter
from collections.abc import Sequence
from copy import deepcopy
from typing import get_type_hints

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORefT
from fastcs.attributes import Attribute, AttrR, AttrW
from fastcs.datatypes import T
from fastcs.tracer import Tracer


class BaseController(Tracer):
    """Base class for controller."""

    #: Attributes passed from the device at runtime.
    attributes: dict[str, Attribute]

    description: str | None = None

    def __init__(
        self,
        path: list[str] | None = None,
        description: str | None = None,
        ios: Sequence[AttributeIO[T, AttributeIORefT]] | None = None,
    ) -> None:
        super().__init__()

        if (
            description is not None
        ):  # Use the argument over the one class defined description.
            self.description = description

        if not hasattr(self, "attributes"):
            self.attributes = {}
        self._path: list[str] = path or []
        self.__sub_controller_tree: dict[str, Controller] = {}

        self._bind_attrs()

        ios = ios or []
        self._attribute_ref_io_map = {io.ref_type: io for io in ios}
        self._validate_io(ios)

    async def initialise(self):
        """Hook to dynamically add attributes before building the API"""
        pass

    def connect_attribute_ios(self) -> None:
        """Connect ``Attribute`` callbacks to ``AttributeIO``s"""
        for attr in self.attributes.values():
            ref = attr.io_ref if attr.has_io_ref() else None
            if ref is None:
                continue

            io = self._attribute_ref_io_map.get(type(ref))
            if io is None:
                raise ValueError(
                    f"{self.__class__.__name__} does not have an AttributeIO "
                    f"to handle {attr.io_ref.__class__.__name__}"
                )

            if isinstance(attr, AttrW):
                attr.set_on_put_callback(io.send)
            if isinstance(attr, AttrR):
                attr.set_update_callback(io.update)

        for controller in self.sub_controllers.values():
            controller.connect_attribute_ios()

    @property
    def path(self) -> list[str]:
        """Path prefix of attributes, recursively including parent Controllers."""
        return self._path

    def set_path(self, path: list[str]):
        if self._path:
            raise ValueError(f"sub controller is already registered under {self.path}")

        self._path = path
        for attribute in self.attributes.values():
            attribute.set_path(path)

    def _bind_attrs(self) -> None:
        """Search for `Attributes` and `Methods` to bind them to this instance.

        This method will search the attributes of this controller class to bind them to
        this specific instance. For `Attribute`s, this is just a case of copying and
        re-assigning to `self` to make it unique across multiple instances of this
        controller class. For `Method`s, this requires creating a bound method from a
        class method and a controller instance, so that it can be called from any
        context with the controller instance passed as the `self` argument.

        """
        # Lazy import to avoid circular references
        from fastcs.cs_methods import UnboundCommand, UnboundScan

        # Using a dictionary instead of a set to maintain order.
        class_dir = {key: None for key in dir(type(self)) if not key.startswith("_")}
        class_type_hints = {
            key: value
            for key, value in get_type_hints(type(self)).items()
            if not key.startswith("_")
        }

        for attr_name in {**class_dir, **class_type_hints}:
            if attr_name == "root_attribute":
                continue

            attr = getattr(self, attr_name, None)
            if isinstance(attr, Attribute):
                setattr(self, attr_name, deepcopy(attr))
            elif isinstance(attr, UnboundScan | UnboundCommand):
                setattr(self, attr_name, attr.bind(self))

    def _validate_io(self, ios: Sequence[AttributeIO[T, AttributeIORefT]]):
        """Validate that there is exactly one AttributeIO class registered to the
        controller for each type of AttributeIORef belonging to the attributes of the
        controller"""
        for ref_type, count in Counter([io.ref_type for io in ios]).items():
            if count > 1:
                raise RuntimeError(
                    f"More than one AttributeIO class handles {ref_type.__name__}"
                )

    def add_attribute(self, name, attribute: Attribute):
        if name in self.attributes and attribute is not self.attributes[name]:
            raise ValueError(
                f"Cannot add attribute {name}. "
                f"Controller {self} has has existing attribute {name}"
            )
        elif name in self.__sub_controller_tree.keys():
            raise ValueError(
                f"Cannot add attribute {name}. "
                f"Controller {self} has existing sub controller {name}"
            )

        attribute.set_name(name)
        attribute.set_path(self.path)
        self.attributes[name] = attribute
        super().__setattr__(name, attribute)

    def add_sub_controller(self, name: str, sub_controller: Controller):
        if name in self.__sub_controller_tree.keys():
            raise ValueError(
                f"Cannot add sub controller {name}. "
                f"Controller {self} has existing sub controller {name}"
            )
        elif name in self.attributes:
            raise ValueError(
                f"Cannot add sub controller {name}. "
                f"Controller {self} has existing attribute {name}"
            )

        sub_controller.set_path(self.path + [name])
        self.__sub_controller_tree[name] = sub_controller
        super().__setattr__(name, sub_controller)

        if isinstance(sub_controller.root_attribute, Attribute):
            self.attributes[name] = sub_controller.root_attribute

    @property
    def sub_controllers(self) -> dict[str, Controller]:
        return self.__sub_controller_tree

    def __repr__(self):
        name = self.__class__.__name__
        path = ".".join(self.path) or None
        sub_controllers = list(self.sub_controllers.keys()) or None

        return f"{name}(path={path}, sub_controllers={sub_controllers})"

    def __setattr__(self, name, value):
        if isinstance(value, Attribute):
            self.add_attribute(name, value)
        elif isinstance(value, Controller):
            self.add_sub_controller(name, value)
        else:
            super().__setattr__(name, value)


class Controller(BaseController):
    """Top-level controller for a device.

    This is the primary class for implementing device support in FastCS. Instances of
    this class can be loaded into a FastCS to expose its ``Attribute``s to the transport
    layer, which can then perform a specific function with the set of ``Attributes``,
    such as generating a UI or creating parameters for a control system.
    """

    root_attribute: Attribute | None = None

    def __init__(
        self,
        description: str | None = None,
        ios: Sequence[AttributeIO[T, AttributeIORefT]] | None = None,
    ) -> None:
        super().__init__(description=description, ios=ios)

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from typing import Any, Generic

from fastcs.attribute_io_ref import AttributeIORefT
from fastcs.datatypes import ATTRIBUTE_TYPES, DataType, T
from fastcs.logging import logger as _logger
from fastcs.tracer import Tracer

ONCE = float("inf")
"""Special value to indicate that an attribute should be updated once on start up."""

logger = _logger.bind(logger_name=__name__)


class Attribute(Generic[T, AttributeIORefT], Tracer):
    """Base FastCS attribute.

    Instances of this class added to a ``Controller`` will be used by the FastCS class.
    """

    def __init__(
        self,
        datatype: DataType[T],
        io_ref: AttributeIORefT | None = None,
        group: str | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__()

        assert issubclass(datatype.dtype, ATTRIBUTE_TYPES), (
            f"Attr type must be one of {ATTRIBUTE_TYPES}, "
            "received type {datatype.dtype}"
        )
        self._io_ref = io_ref
        self._datatype: DataType[T] = datatype
        self._group = group
        self.enabled = True
        self.description = description

        # A callback to use when setting the datatype to a different value, for example
        # changing the units on an int.
        self._update_datatype_callbacks: list[Callable[[DataType[T]], None]] = []

        # Path and name to be filled in by Controller it is bound to
        self._name = ""
        self._path = []

    @property
    def io_ref(self) -> AttributeIORefT:
        if self._io_ref is None:
            raise RuntimeError(f"{self} has no AttributeIORef")
        return self._io_ref

    def has_io_ref(self):
        return self._io_ref is not None

    @property
    def datatype(self) -> DataType[T]:
        return self._datatype

    @property
    def dtype(self) -> type[T]:
        return self._datatype.dtype

    @property
    def group(self) -> str | None:
        return self._group

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> list[str]:
        return self._path

    def add_update_datatype_callback(
        self, callback: Callable[[DataType[T]], None]
    ) -> None:
        self._update_datatype_callbacks.append(callback)

    def update_datatype(self, datatype: DataType[T]) -> None:
        if not isinstance(self._datatype, type(datatype)):
            raise ValueError(
                f"Attribute datatype must be of type {type(self._datatype)}"
            )
        self._datatype = datatype
        for callback in self._update_datatype_callbacks:
            callback(datatype)

    def set_name(self, name: str):
        if self._name:
            raise RuntimeError(
                f"Attribute is already registered with a controller as {self._name}"
            )

        self._name = name

    def set_path(self, path: list[str]):
        if self._path:
            raise RuntimeError(
                f"Attribute is already registered with a controller at {self._path}"
            )

        self._path = path

    def __repr__(self):
        name = self.__class__.__name__
        path = ".".join(self._path + [self._name]) or None
        datatype = self._datatype.__class__.__name__

        return f"{name}(path={path}, datatype={datatype}, io_ref={self._io_ref})"


AttrIOUpdateCallback = Callable[["AttrR[T, Any]"], Awaitable[None]]
"""An AttributeIO callback that takes an AttrR and updates its value"""
AttrUpdateCallback = Callable[[], Awaitable[None]]
"""A callback to be called periodically to update an attribute"""
AttrOnUpdateCallback = Callable[[T], Awaitable[None]]
"""A callback to be called when the value of the attribute is updated"""


class AttrR(Attribute[T, AttributeIORefT]):
    """A read-only ``Attribute``."""

    def __init__(
        self,
        datatype: DataType[T],
        io_ref: AttributeIORefT | None = None,
        group: str | None = None,
        initial_value: T | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__(datatype, io_ref, group, description=description)
        self._value: T = (
            datatype.initial_value if initial_value is None else initial_value
        )
        self._update_callback: AttrIOUpdateCallback[T] | None = None
        """Callback to update the value of the attribute with an IO to the source"""
        self._on_update_callbacks: list[AttrOnUpdateCallback[T]] | None = None
        """Callbacks to publish changes to the value of the attribute"""

    def get(self) -> T:
        """Get the cached value of the attribute."""
        return self._value

    async def update(self, value: T) -> None:
        """Update the value of the attibute

        This sets the cached value of the attribute presented in the API. It should
        generally only be called from an IO or a controller that is updating the value
        from some underlying source.

        To request a change to the setpoint of the attribute, use the ``put`` method,
        which will attempt to apply the change to the underlying source.

        """
        self.log_event("Attribute set", attribute=self, value=value)

        self._value = self._datatype.validate(value)

        if self._on_update_callbacks is not None:
            try:
                await asyncio.gather(
                    *[cb(self._value) for cb in self._on_update_callbacks]
                )
            except Exception as e:
                logger.opt(exception=e).error(
                    "On update callback failed", attribute=self, value=value
                )

    def add_on_update_callback(self, callback: AttrOnUpdateCallback[T]) -> None:
        """Add a callback to be called when the value of the attribute is updated

        The callback will be called with the updated value.

        """
        if self._on_update_callbacks is None:
            self._on_update_callbacks = []
        self._on_update_callbacks.append(callback)

    def set_update_callback(self, callback: AttrIOUpdateCallback[T]):
        """Set the callback to update the value of the attribute from the source

        The callback will be converted to an async task and called periodically.

        """
        if self._update_callback is not None:
            raise RuntimeError("Attribute already has an IO update callback")

        self._update_callback = callback

    def bind_update_callback(self) -> AttrUpdateCallback:
        """Bind self into the registered IO update callback"""
        if self._update_callback is None:
            raise RuntimeError("Attribute has no update callback")
        else:
            update_callback = self._update_callback

        async def update_attribute():
            try:
                self.log_event("Update attribute", topic=self)
                await update_callback(self)
            except Exception as e:
                logger.opt(exception=e).error("Update loop failed", attribute=self)
                raise

        return update_attribute


AttrOnPutCallback = Callable[["AttrW[T, Any]", T], Awaitable[None]]
"""Callbacks to be called when the setpoint of an attribute is changed"""
AttrSyncSetpointCallback = Callable[[T], Awaitable[None]]
"""Callbacks to be called when the setpoint of an attribute is changed"""


class AttrW(Attribute[T, AttributeIORefT]):
    """A write-only ``Attribute``."""

    def __init__(
        self,
        datatype: DataType[T],
        io_ref: AttributeIORefT | None = None,
        group: str | None = None,
        description: str | None = None,
    ) -> None:
        super().__init__(
            datatype,  # type: ignore
            io_ref,
            group,
            description=description,
        )
        self._on_put_callback: AttrOnPutCallback[T] | None = None
        """Callback to action a change to the setpoint of the attribute"""
        self._sync_setpoint_callbacks: list[AttrSyncSetpointCallback[T]] = []
        """Callbacks to publish changes to the setpoint of the attribute"""

    async def put(self, setpoint: T, sync_setpoint: bool = False) -> None:
        """Set the setpoint of the attribute

        This should be called by clients to the attribute such as transports to apply a
        change to the attribute. The ``_on_put_callback`` will be called with this new
        setpoint, which may or may not take effect depending on the validity of the new
        value. For example, if the attribute has an IO to some device, the value might
        be rejected.

        To directly change the value of the attribute, for example from an update loop
        that has read a new value from some underlying source, call the ``set`` method.

        """
        setpoint = self._datatype.validate(setpoint)
        if self._on_put_callback is not None:
            try:
                await self._on_put_callback(self, setpoint)
            except Exception as e:
                logger.opt(exception=e).error(
                    "Put failed", attribute=self, setpoint=setpoint
                )

        if sync_setpoint:
            try:
                await self._call_sync_setpoint_callbacks(setpoint)
            except Exception as e:
                logger.opt(exception=e).error(
                    "Sync setpoint failed", attribute=self, setpoint=setpoint
                )

    async def _call_sync_setpoint_callbacks(self, setpoint: T) -> None:
        if self._sync_setpoint_callbacks:
            await asyncio.gather(
                *[cb(setpoint) for cb in self._sync_setpoint_callbacks]
            )

    def set_on_put_callback(self, callback: AttrOnPutCallback[T]) -> None:
        """Set the callback to call when the setpoint is changed

        The callback will be called with the attribute and the new setpoint.

        """
        if self._on_put_callback is not None:
            raise RuntimeError("Attribute already has an on put callback")

        self._on_put_callback = callback

    def add_sync_setpoint_callback(self, callback: AttrSyncSetpointCallback[T]) -> None:
        """Add a callback to publish changes to the setpoint of the attribute

        The callback will be called with the new setpoint.

        """
        self._sync_setpoint_callbacks.append(callback)


class AttrRW(AttrR[T, AttributeIORefT], AttrW[T, AttributeIORefT]):
    """A read-write ``Attribute``."""

    def __init__(
        self,
        datatype: DataType[T],
        io_ref: AttributeIORefT | None = None,
        group: str | None = None,
        initial_value: T | None = None,
        description: str | None = None,
    ):
        super().__init__(datatype, io_ref, group, initial_value, description)

        self._setpoint_initialised = False

        if io_ref is None:
            self.set_on_put_callback(self._internal_update)

    async def _internal_update(self, attr: AttrW[T, AttributeIORefT], value: T):
        """Update value directly when Attribute has no IO"""
        assert attr is self
        await self.update(value)

    async def update(self, value: T):
        await super().update(value)

        if not self._setpoint_initialised:
            await self._call_sync_setpoint_callbacks(value)
            self._setpoint_initialised = True



from dataclasses import dataclass
from enum import Enum
from pathlib import Path


@dataclass
class EpicsDocsOptions:
    """Docs options for EPICS."""

    path: Path = Path(".")
    depth: int | None = None


class EpicsGUIFormat(Enum):
    """The format of an EPICS GUI."""

    bob = ".bob"
    edl = ".edl"


@dataclass
class EpicsGUIOptions:
    """Epics GUI options for use in both CA and PVA transports."""

    output_path: Path = Path(".") / "output.bob"
    file_format: EpicsGUIFormat = EpicsGUIFormat.bob
    title: str = "Simple Device"


@dataclass
class EpicsIOCOptions:
    """Epics IOC options for use in both CA and PVA transports."""

    pv_prefix: str = "MY-DEVICE-PREFIX"



import json
import logging
import traceback
from os import _exit
from random import uniform

import numpy as np
import numpy.typing as npt
import pydantic.v1.dataclasses
from tickit.adapters.io import TcpIo
from tickit.adapters.specifications.regex_command import RegexCommand
from tickit.adapters.tcp import CommandAdapter
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat
from typing_extensions import TypedDict

LOGGER = logging.getLogger(__name__)


def handle_exceptions(func):
    """Log and exit if the wrapped function raises an Exception.

    Tickit does not properly handle exceptions in e.g. update() methods. This wrapper is
    used to ensure we see the errors as soon as they occur.
    """

    def wrapped_func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:  # noqa
            logging.error(traceback.format_exc())
            _exit(-1)

    return wrapped_func


class TempControllerDevice(Device):
    class Inputs(TypedDict):
        flux: float

    class Outputs(TypedDict):
        flux: float

    def __init__(
        self,
        num_ramp_controllers: int,
        default_start: float,
        default_end: float,
    ) -> None:
        self._num = num_ramp_controllers
        self._start = np.full(num_ramp_controllers, default_start, dtype=int)
        self._end = np.full(num_ramp_controllers, default_end, dtype=int)
        self._target = np.zeros(num_ramp_controllers, dtype=float)
        self._actual = np.zeros(num_ramp_controllers, dtype=float)
        self._enabled = np.full(num_ramp_controllers, 0, dtype=int)
        self._start_times = np.full(num_ramp_controllers, -1, dtype=int)

        self._ramp_rate: float = 1  # 1 unit/s

        self.initialised = False

    def get_target(self, index: int):
        return self._target[index]

    def get_actual(self, index: int):
        return self._actual[index]

    def get_start(self, index: int):
        return self._start[index]

    def set_start(self, index: int, value: int):
        self._start[index] = value
        LOGGER.info(f"Set ramp {index} start to {value}")

    def get_end(self, index: int):
        return self._end[index]

    def set_end(self, index: int, value: int):
        self._end[index] = value
        LOGGER.info(f"Set ramp {index} end to {value}")

    def set_enabled(self, index: int, value: int):
        if value:
            self._target[index] = float(self._start[index])
            self._start_times[index] = -1
            self._enabled[index] = 1
            LOGGER.info(f"Started ramp {index}")
        else:
            self._enabled[index] = 0
            LOGGER.info(f"Stopped ramp {index}")

    def get_enabled(self, index: int):
        return self._enabled[index]

    def get_ramp_rate(self):
        return self._ramp_rate

    def set_ramp_rate(self, rate: float):
        self._ramp_rate = rate
        LOGGER.info(f"Set ramp rate to {rate}")

    def get_voltages(self):
        return [self.get_actual(i) / 17 for i in range(0, self._num)]

    def get_power(self):
        return sum(self.get_voltages() * 50)

    def ramp(
        self,
        periods: npt.NDArray,
    ):
        can_ramp = np.logical_and(self._enabled, self._target < self._end)
        max_temps = self._target.copy()
        max_temps[can_ramp] += self._ramp_rate * (periods[can_ramp] / 1e9)
        self._target = np.minimum(max_temps, self._end)
        self._enabled[self._target >= self._end] = 0

        self._actual = [
            float(self.get_target(i) + uniform(-0.5, 0.5)) if self.get_enabled(i) else 0
            for i in range(0, self._num)
        ]

    @handle_exceptions
    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        if not self.initialised:
            LOGGER.info("Temperature controller running")
            self.initialised = True

        self._start_times[self._start_times == -1] = int(time)
        periods = np.full(self._num, int(time)) - self._start_times
        self.ramp(periods)
        self._start_times = np.full(self._num, int(time))

        LOGGER.info(
            f"Target Temperatures: {', '.join(f'{float(t):.3f}' for t in self._target)}"
        )
        LOGGER.info(
            f"Actual Temperatures: {', '.join(f'{float(t):.3f}' for t in self._actual)}"
        )

        call_at = SimTime(int(time) + int(1e8)) if np.any(self._enabled) else None
        return DeviceUpdate(TempControllerDevice.Outputs(flux=inputs["flux"]), call_at)

    def get_commands(self):
        return json.dumps(
            {
                "Device ID": {"command": "ID", "type": "str", "access_mode": "r"},
                "Power": {"command": "P", "type": "float", "access_mode": "rw"},
                "Ramp Rate": {"command": "R", "type": "float", "access_mode": "rw"},
                "Ramps": [
                    {
                        "Start": {
                            "command": f"S{idx:02d}",
                            "type": "int",
                            "access_mode": "rw",
                        },
                        "End": {
                            "command": f"E{idx:02d}",
                            "type": "int",
                            "access_mode": "rw",
                        },
                        "Enabled": {
                            "command": f"N{idx:02d}",
                            "type": "int",
                            "access_mode": "rw",
                        },
                        "Target": {
                            "command": f"T{idx:02d}",
                            "type": "float",
                            "access_mode": "rw",
                        },
                        "Actual": {
                            "command": f"A{idx:02d}",
                            "type": "float",
                            "access_mode": "rw",
                        },
                    }
                    for idx in range(1, self._num + 1)
                ],
            }
        )


class TempControllerAdapter(CommandAdapter):
    device: TempControllerDevice
    _byte_format: ByteFormat = ByteFormat(b"%b\r\n")

    def __init__(self, device: TempControllerDevice) -> None:
        super().__init__()
        self.device = device

    def _validate_index(self, index: str) -> int:
        int_index = int(index) - 1
        assert int_index >= 0, "TempController Indices start at 01"
        return int_index

    @RegexCommand(r"ID\?", False, "utf-8")
    async def get_id(self) -> bytes:
        return b"SIMTCONT123"

    @RegexCommand(r"T([0-9][0-9])\?", False, "utf-8")
    async def get_target(self, index: str) -> bytes:
        int_index = self._validate_index(index)
        return str(self.device.get_target(int_index)).encode("utf-8")

    @RegexCommand(r"A([0-9][0-9])\?", False, "utf-8")
    async def get_actual(self, index: str) -> bytes:
        int_index = self._validate_index(index)
        return str(self.device.get_actual(int_index)).encode("utf-8")

    @RegexCommand(r"N([0-9][0-9])\?", False, "utf-8")
    async def get_enabled(self, index: str) -> bytes:
        int_index = self._validate_index(index)
        return str(self.device.get_enabled(int_index)).encode("utf-8")

    @RegexCommand(r"N([0-9][0-9])=([01])", True, "utf-8")
    async def set_enabled(self, index: str, value: int) -> None:
        int_index = self._validate_index(index)
        self.device.set_enabled(int_index, value)

    @RegexCommand(r"S([0-9][0-9])\?", False, "utf-8")
    async def get_start(self, index: str) -> bytes:
        int_index = self._validate_index(index)
        return str(self.device.get_start(int_index)).encode("utf-8")

    @RegexCommand(r"S([0-9][0-9])=(\d+\.?\d*)", True, "utf-8")
    async def set_start(self, index: str, value: str) -> None:
        int_index = self._validate_index(index)
        self.device.set_start(int_index, int(value))

    @RegexCommand(r"E([0-9][0-9])\?", False, "utf-8")
    async def get_end(self, index: str) -> bytes:
        int_index = self._validate_index(index)
        return str(self.device.get_end(int_index)).encode("utf-8")

    @RegexCommand(r"E([0-9][0-9])=(\d+\.?\d*)", True, "utf-8")
    async def set_end(self, index: str, value: str) -> None:
        int_index = self._validate_index(index)
        self.device.set_end(int_index, int(value))

    @RegexCommand(r"R\?", False, "utf-8")
    async def get_ramp_rate(self) -> bytes:
        return str(self.device.get_ramp_rate()).encode("utf-8")

    @RegexCommand(r"R=(\d+\.?\d*)", True, "utf-8")
    async def set_ramp_rate(self, value: str) -> None:
        self.device.set_ramp_rate(float(value))

    @RegexCommand(r"V\?", False, "utf-8")
    async def get_voltages(self) -> bytes:
        return str(self.device.get_voltages()).encode("utf-8")

    @RegexCommand(r"P\?", False, "utf-8")
    async def get_power(self) -> bytes:
        return str(self.device.get_power()).encode("utf-8")

    @RegexCommand(r"API\?", False, "utf-8")
    async def get_commands(self) -> bytes:
        return str(self.device.get_commands()).encode("utf-8")

    @RegexCommand(r"\w*", False, "utf-8")
    async def ignore_whitespace(self) -> None:
        pass


# https://github.com/DiamondLightSource/tickit/issues/212
@pydantic.v1.dataclasses.dataclass
class TempController(ComponentConfig):  # type: ignore
    num_ramp_controllers: int
    default_start: float = 0
    default_end: float = 50
    host: str = "localhost"
    port: int = 25565

    def __call__(self) -> Component:
        device = TempControllerDevice(
            num_ramp_controllers=self.num_ramp_controllers,
            default_start=self.default_start,
            default_end=self.default_end,
        )
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=[
                AdapterContainer(
                    TempControllerAdapter(device),
                    TcpIo(self.host, self.port),
                )
            ],
        )



"""Top level API.

.. data:: __version__
    :type: str

    Version number as calculated by https://github.com/pypa/setuptools_scm
"""

from . import attributes as attributes
from . import controller as controller
from . import cs_methods as cs_methods
from . import datatypes as datatypes
from . import transport as transport
from ._version import __version__ as __version__
from .control_system import FastCS as FastCS



from .ip_connection import IPConnection as IPConnection
from .ip_connection import IPConnectionSettings as IPConnectionSettings
from .ip_connection import StreamConnection as StreamConnection
from .serial_connection import SerialConnection as SerialConnection
from .serial_connection import SerialConnectionSettings as SerialConnectionSettings



import numpy as np

from fastcs.datatypes import Bool, DataType, Enum, Float, Int, String, T, Waveform

REST_ALLOWED_DATATYPES = (Bool, DataType, Enum, Float, Int, String)


def convert_datatype(datatype: DataType[T]) -> type:
    """Converts a datatype to a rest serialisable type."""
    match datatype:
        case Waveform():
            return list
        case _:
            return datatype.dtype


def cast_to_rest_type(datatype: DataType[T], value: T) -> object:
    """Casts from an attribute value to a rest value."""
    match datatype:
        case Waveform():
            return value.tolist()
        case datatype if issubclass(type(datatype), REST_ALLOWED_DATATYPES):
            return datatype.validate(value)
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")


def cast_from_rest_type(datatype: DataType[T], value: object) -> T:
    """Casts from a rest value to an attribute datatype."""
    match datatype:
        case Waveform():
            return datatype.validate(np.array(value, dtype=datatype.array_dtype))
        case datatype if issubclass(type(datatype), REST_ALLOWED_DATATYPES):
            return datatype.validate(value)  # type: ignore
        case _:
            raise ValueError(f"Unsupported datatype {datatype}")



import asyncio
import enum
import json
from dataclasses import KW_ONLY, dataclass
from typing import TypeVar

from fastcs.attribute_io import AttributeIO
from fastcs.attribute_io_ref import AttributeIORef
from fastcs.attributes import AttrR, AttrRW, AttrW
from fastcs.connections import IPConnection, IPConnectionSettings
from fastcs.controller import Controller
from fastcs.datatypes import Enum, Float, Int
from fastcs.wrappers import command, scan

NumberT = TypeVar("NumberT", int, float)


class OnOffEnum(enum.StrEnum):
    Off = "0"
    On = "1"


@dataclass
class TemperatureControllerSettings:
    num_ramp_controllers: int
    ip_settings: IPConnectionSettings


@dataclass
class TemperatureControllerAttributeIORef(AttributeIORef):
    name: str
    _: KW_ONLY
    update_period: float | None = 0.2


class TemperatureControllerAttributeIO(
    AttributeIO[NumberT, TemperatureControllerAttributeIORef]
):
    def __init__(self, connection: IPConnection, suffix: str):
        super().__init__()

        self._connection = connection
        self.suffix = suffix

    async def send(
        self, attr: AttrW[NumberT, TemperatureControllerAttributeIORef], value: NumberT
    ) -> None:
        command = f"{attr.io_ref.name}{self.suffix}={attr.dtype(value)}"
        await self._connection.send_command(f"{command}\r\n")
        self.log_event("Send command for attribute", topic=attr, command=command)

    async def update(
        self, attr: AttrR[NumberT, TemperatureControllerAttributeIORef]
    ) -> None:
        query = f"{attr.io_ref.name}{self.suffix}?"
        response = await self._connection.send_query(f"{query}\r\n")
        response = response.strip("\r\n")
        self.log_event(
            "Query for attribute",
            topic=attr,
            query=query,
            response=response,
        )

        await attr.update(attr.dtype(response))


class TemperatureController(Controller):
    ramp_rate = AttrRW(Float(), io_ref=TemperatureControllerAttributeIORef(name="R"))
    power = AttrR(Float(), io_ref=TemperatureControllerAttributeIORef(name="P"))

    def __init__(self, settings: TemperatureControllerSettings) -> None:
        self.connection = IPConnection()
        self.suffix = ""
        super().__init__(
            ios=[TemperatureControllerAttributeIO(self.connection, self.suffix)]
        )

        self._settings = settings

        self._ramp_controllers: list[TemperatureRampController] = []
        for index in range(1, settings.num_ramp_controllers + 1):
            controller = TemperatureRampController(index, self.connection)
            self._ramp_controllers.append(controller)
            self.add_sub_controller(f"R{index}", controller)

    @command()
    async def cancel_all(self) -> None:
        for rc in self._ramp_controllers:
            await rc.enabled.put(OnOffEnum.Off, sync_setpoint=True)
            # TODO: The requests all get concatenated and the sim doesn't handle it
            await asyncio.sleep(0.1)

    async def connect(self) -> None:
        await self.connection.connect(self._settings.ip_settings)

    async def close(self) -> None:
        await self.connection.close()

    @scan(0.1)
    async def update_voltages(self):
        query = "V?"
        voltages = json.loads(
            (await self.connection.send_query(f"{query}\r\n")).strip("\r\n")
        )
        for index, controller in enumerate(self._ramp_controllers):
            self.log_event(
                "Update voltages",
                topic=controller.voltage,
                query=query,
                response=voltages,
            )
            await controller.voltage.update(float(voltages[index]))


class TemperatureRampController(Controller):
    start = AttrRW(Int(), io_ref=TemperatureControllerAttributeIORef(name="S"))
    end = AttrRW(Int(), io_ref=TemperatureControllerAttributeIORef(name="E"))
    enabled = AttrRW(
        Enum(OnOffEnum), io_ref=TemperatureControllerAttributeIORef(name="N")
    )
    target = AttrR(Float(prec=3), io_ref=TemperatureControllerAttributeIORef(name="T"))
    actual = AttrR(Float(prec=3), io_ref=TemperatureControllerAttributeIORef(name="A"))
    voltage = AttrR(Float(prec=3))

    def __init__(self, index: int, conn: IPConnection) -> None:
        suffix = f"{index:02d}"
        super().__init__(
            f"Ramp{suffix}", ios=[TemperatureControllerAttributeIO(conn, suffix)]
        )
        self.connection = conn



from dataclasses import dataclass
from typing import NewType


@dataclass
class GraylogEndpoint:
    """Server and port for a graylog instance."""

    host: str
    port: int

    @classmethod
    def parse_graylog_endpoint(cls, endpoint: str) -> "GraylogEndpoint":
        try:
            host, port = endpoint.split(":")
            port = int(port)
        except Exception as e:
            raise ValueError(
                "Invalid graylog endpoint. Expected '<host>:<port>'."
            ) from e

        return cls(host, port)


GraylogEnvFields = NewType("GraylogEnvFields", dict[str, str])
"""Fields to add to graylog messages from environment variables."""


def parse_graylog_env_fields(comma_separated_fields: str) -> GraylogEnvFields:
    try:
        return GraylogEnvFields(_parse_graylog_env_fields(comma_separated_fields))
    except Exception as e:
        raise ValueError(
            "Failed to parse fields. Expected '<field_name>:<env_name>,...'"
        ) from e


GraylogStaticFields = NewType("GraylogStaticFields", dict[str, str])
"""Fields to add to graylog messages with static values."""


def parse_graylog_static_fields(comma_separated_fields: str) -> GraylogStaticFields:
    try:
        return GraylogStaticFields(_parse_graylog_env_fields(comma_separated_fields))
    except Exception as e:
        raise ValueError("Failed to parse fields. Expected '<key>:<value>,...'") from e


def _parse_graylog_env_fields(comma_separated_fields: str) -> dict[str, str]:
    return {
        field: env
        for pair in comma_separated_fields.split(",")
        for field, env in [pair.split(":")]
    }
